#include <stdio.h> 
#include <stdlib.h> 
#include <string.h>

#include <glib-object.h>
#include <json-glib/json-glib.h>
#include <json-glib/json-gobject.h>

#include "KAIN2.H"
  
int main () 
{ 
    FILE *dataFile; 
	char *filePath = "cliff1.drm";
	char *fileBuffer;
    struct Level *levelData;
	size_t fileSize;
	unsigned long dataOffset;
	unsigned long numPointers;
	unsigned long *pointerAddresses;
	int i;
	
	// JSON serialization
	gchar *jsonData;
	gsize jsonLength;
      
    // Open person.dat for reading 
    dataFile = fopen(filePath, "r"); 
    if (dataFile == NULL) 
    { 
        fprintf(stderr, "Could not open file '%s'\n", filePath); 
        exit(1); 
    }
	printf("Level struct size: %i\n", sizeof(struct Level)); 
	
	fseek(dataFile, 0L, SEEK_END);
	fileSize = ftell(dataFile);
	rewind(dataFile);
	printf("Input file size: %i\n", fileSize);
	fileBuffer = (char*)malloc(fileSize * sizeof(char) * 1);
	if (fileBuffer == NULL)
	{
		fprintf(stderr, "Couldn't allocate memory for file buffer\n"); 
        exit(1); 
	}
	fread(&numPointers, sizeof(unsigned long), 1, dataFile);	
    dataOffset = ((numPointers >> 9) << 11) + 0x00000800;
	printf("Data offset: %i\n", dataOffset);
	printf("Number of pointers: %i\n", numPointers);
	
	// read the pointer offsets into a separate buffer
	pointerAddresses = (unsigned long*)malloc(numPointers * sizeof(unsigned long) * 1);
	if (pointerAddresses == NULL)
	{
		fprintf(stderr, "Couldn't allocate memory for pointer addresses\n"); 
        exit(1); 
	}
	printf("Reading pointer offsets\n");
	fseek(dataFile, 4, SEEK_SET);
	fread(pointerAddresses, sizeof(unsigned long), numPointers, dataFile);
	
	// read the actual file data
	printf("Reading file data\n");
	fseek(dataFile, dataOffset, SEEK_SET);
	
    //fread(&levelData, sizeof(struct Level), 1, dataFile);
	fread(fileBuffer, fileSize - dataOffset, 1, dataFile);
    fclose(dataFile);
	printf("Done reading '%s'\n", filePath); 
	printf("*fileBuffer address: %p\n", fileBuffer);
	
	// update the pointers
	for (i = 0; i < numPointers; i++)
	{
		unsigned long pointerAddress = pointerAddresses[i];
		unsigned long existingValue = *((unsigned long *) &fileBuffer[pointerAddress]);
		unsigned long newValue = fileBuffer + existingValue;
		printf("pointerAddress: %i\n", pointerAddress);
		printf("\texistingValue: %x\n", existingValue);
		printf("\tnewValue: %x\n", newValue);
		//memcpy(fileBuffer[pointerAddress], &newValue, sizeof(newValue));
		memcpy(fileBuffer + pointerAddress, (char*)&newValue, sizeof(newValue));
	}
	
	// populate the Level struct
	//levelData = (struct Level*)malloc(sizeof(struct Level));
	//memcpy(levelData, fileBuffer, sizeof(struct Level));
	levelData = (struct Level*)fileBuffer;

	printf("*terrain: %p\n", levelData->terrain);
	printf("terrain.vplLength: %i\n", levelData->terrain->vplLength); 
	printf("terrain.numIntros: %i\n", levelData->terrain->numIntros);

	printf("numVMObjects: %i\n", levelData->numVMObjects); 
	for (i = 0; i < levelData->numVMObjects; i++)
	{
		printf("\t%s\n", levelData->vmobjectList[i].name);
	}
	
	printf("morphLastStep: %i\n", levelData->morphLastStep); 
    printf("worldname: %s\n", levelData->worldName);
	printf("numberBrotherUnits: %i\n", levelData->numberBrotherUnits); 
	
	jsonData = json_gobject_to_data(G_OBJECT(levelData), &jsonLength);
	g_print("JSON:\n%s\n", jsonData);
	g_free(jsonData);
	g_object_unref(levelData);
  
    return 0; 
} 