﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Recombobulator
{
    static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main(string[] args)
        {
            bool guiMode = true;
            if (args.Length > 0)
            {
                guiMode = false;
            }

            if (guiMode)
            {
                Application.EnableVisualStyles();
                Application.SetCompatibleTextRenderingDefault(false);
                Application.Run(new MainForm());
            }
            else
            {
                string inputFilePath = "";
                Recombobulator.SR1_File.SR1FileType fileType = SR1_File.SR1FileType.AreaRelease;
                try
                {                    
                    string outputFilePath = "";

                    ArrayList arguments = new ArrayList();
                    for (int i = 0; i < args.Length; i++)
                    {
                        // if the argument is one with a second component, catch that now
                        string nextArg = "";
                        switch (args[i])
                        {
                            case "--input":
                                nextArg = args[i + 1];
                                inputFilePath = Path.GetFullPath(nextArg);
                                i++;
                                break;
                            case "--output":
                                nextArg = args[i + 1];
                                outputFilePath = Path.GetFullPath(nextArg);
                                i++;
                                break;
                            case "--filetype":
                                nextArg = args[i + 1].ToLower().Trim();
                                switch(nextArg)
                                {
                                    case "alpha19990123":
                                        fileType = SR1_File.SR1FileType.AreaAlpha19990123;
                                        break;
                                    case "alpha19990204":
                                        fileType = SR1_File.SR1FileType.AreaAlpha19990204;
                                        break;
                                    case "alpha19990216":
                                        fileType = SR1_File.SR1FileType.AreaAlpha19990216;
                                        break;
                                    case "beta19990512":
                                        fileType = SR1_File.SR1FileType.AreaBeta19990512;
                                        break;
                                    case "beta":
                                        fileType = SR1_File.SR1FileType.AreaBeta;
                                        break;
                                    case "release":
                                        fileType = SR1_File.SR1FileType.AreaRelease;
                                        break;
                                }
                                i++;
                                break;
                            default:
                                arguments.Add(args[i]);
                                break;
                        }
                    }

                    ArrayList positionalArgs = new ArrayList();

                    foreach (string arg in arguments)
                    {
                        bool handled = false;
                        if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                        {
                            Console.WriteLine("Run Recombobulator with no arguments to access the GUI mode. Use --input INPUT_FILE and --output OUTPUT_FILE to convert in batch mode");
                            Environment.Exit(0);
                        }
                        if (!handled)
                        {
                            positionalArgs.Add(arg);
                        }
                    }
                    if (inputFilePath == "")
                    {
                        Console.WriteLine("--input path is required");
                        Environment.Exit(1);
                    }
                    if (outputFilePath == "")
                    {
                        Console.WriteLine("--output path is required");
                        Environment.Exit(1);
                    }

                    SR1_File inFile = new SR1_File();

                    inFile.Import(inputFilePath, fileType);

                    string jsonData = Newtonsoft.Json.JsonConvert.SerializeObject(inFile.GetProperties(), Newtonsoft.Json.Formatting.Indented);
                    //using non-indented formatting saves enough memory to make conversion possible for some problem areas that cause out-of-memory exceptions otherwise
                    //string jsonData = Newtonsoft.Json.JsonConvert.SerializeObject(inFile.GetProperties(), Newtonsoft.Json.Formatting.None);

                    FileStream outStream = new FileStream(outputFilePath, FileMode.Create);
                    StreamWriter outWriter = new StreamWriter(outStream);
                    outWriter.Write(jsonData);
                    outWriter.Close();
                    outStream.Close();
                }
                catch (Exception ex)
                {
                    Console.WriteLine(string.Format("Error processing '{0}': {1}", inputFilePath, ex.Message));
                    Console.WriteLine(ex.StackTrace);
                    Environment.Exit(1);
                }
            }
        }
    }
}
