﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Recombobulator
{
    class ObjectNameList : SR1_Structure
    {
        private List<SR1_PrimitiveArray<char>> _List = new List<SR1_PrimitiveArray<char>>();

        public ObjectNameList()
        {
        }

        public override TreeList.Node[] CreatePropertyNodes()
        {
            string memberTypeName = GetPrimitiveTypeName(TypeCode.Char);
            memberTypeName += "[";
            if (_List.Count > 0)
            {
                memberTypeName += _List.Count;
            }
            memberTypeName += "][16]";
            string memberName = "names";
            string memberValue = "{...}";

            TreeList.Node[] memberNodes = new TreeList.Node[1];
            memberNodes[0] = new TreeList.Node(new string[] { "", memberTypeName, memberName, memberValue });

            if (_List.Count > 0)
            {
                TreeList.Node[] elementNodes = new TreeList.Node[_List.Count];
                for (int i = 0; i < _List.Count; i++)
                {
                    string elementTypeName = _List[i].GetTypeName(true);
                    string elementName = "";
                    string elementValue = _List[i].ToString();
                    elementName = "";

                    elementNodes[i] = new TreeList.Node(new string[] { "", elementTypeName, elementName, elementValue });
                    elementNodes[i].Nodes.AddRange(_List[i].CreatePropertyNodes());
                }

                memberNodes[0].Nodes.AddRange(elementNodes);
            }

            return memberNodes;
        }

        public override void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            while (true)
            {
                byte next = reader.ReadByte();
                if (next == 0xFF)
                {
                    break;
                }

                reader.BaseStream.Position--;

                SR1_PrimitiveArray<char> name = new SR1_PrimitiveArray<char>(16);
                name.Read(dictionary, reader, true, tracker);
                _List.Add(name);
            }

            reader.BaseStream.Position += 3;

            End = (uint)reader.BaseStream.Position;
        }
    }
}
