﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Recombobulator.SR1Structures;

namespace Recombobulator
{
    class SR1_File
    {
        public enum SR1FileType
        {
            AreaAlpha19990123,
            AreaAlpha19990204,
            AreaAlpha19990216,
            AreaBeta19990512,
            AreaBeta,
            AreaRelease
        }

        protected SortedDictionary<uint, SR1_Structure> _Dictionary = new SortedDictionary<uint, SR1_Structure>();

        public SortedDictionary<uint, SR1_Structure> Dictionary
        {
            get
            {
                return _Dictionary;
            }
            set
            {
                _Dictionary = value;
            }
        }

        public SR1_File()
        {
        }

        public void Import(string fileName, SR1FileType fileType)
        {
            _Dictionary.Clear();

            FileStream file = new FileStream(fileName, FileMode.Open, FileAccess.Read);
            MemoryStream stream = new MemoryStream((int)file.Length);
            BinaryReader streamReader = new BinaryReader(stream);
            BinaryWriter streamWriter = new BinaryWriter(stream);

            BinaryReader fileReader = new BinaryReader(file, System.Text.Encoding.UTF8);
            ResolvePointers(fileReader, streamWriter);
            fileReader.Close();

            Level level = new Level();
            if (fileType == SR1FileType.AreaAlpha19990123)
            {
                level = new LevelAlpha19990123();
            }
            if (fileType == SR1FileType.AreaAlpha19990204)
            {
                level = new LevelAlpha19990204();
            }
            if (fileType == SR1FileType.AreaAlpha19990216)
            {
                level = new LevelAlpha19990216();
            }
            if (fileType == SR1FileType.AreaBeta19990512)
            {
                level = new Level19990512();
            }
            GoalTracker importTracker = new GoalTracker();
            level.Read(_Dictionary, streamReader, false, importTracker);

            streamReader.Close();
            streamWriter.Close();
        }

        public void Export(string fileName)
        {

        }

        protected void ResolvePointers(BinaryReader reader, BinaryWriter writer)
        {
            UInt32 uDataStart = ((reader.ReadUInt32() >> 9) << 11) + 0x00000800;
            //if (xReader.ReadUInt32() == 0x00000000)
            //{
            //    _asset = Asset.Unit;
            //}
            //else
            //{
            //    _asset = Asset.Object;
            //}

            reader.BaseStream.Position = uDataStart;
            writer.BaseStream.Position = 0;
            reader.BaseStream.CopyTo(writer.BaseStream);
            writer.BaseStream.Position = 0;
        }

        public TreeList.Node[] CreateChunkNodes()
        {
            List<TreeList.Node> nodeList = new List<TreeList.Node>();
            foreach (KeyValuePair<uint, SR1_Structure> entry in _Dictionary)
            {
                uint lookUpID = entry.Key;
                SR1_Structure structure = entry.Value;

                if (lookUpID > 0 && structure.IsMember)
                {
                    continue;
                }

                string typeName = structure.GetTypeName(true);
                string name = structure.GetTypeName(false).ToLower();
                string offset = " (0x" + structure.Start.ToString("X8") + "-" + "0x" + structure.End.ToString("X8") + ")";
                string value = structure.ToString();

                TreeList.Node newNode = new TreeList.Node(new string[] { offset, typeName, name, value });
                newNode.Nodes.AddRange(structure.CreatePropertyNodes());
                nodeList.Add(newNode);
            }

            TreeList.Node dummyNode = new TreeList.Node(new string[] { "", "", "", "" });
            nodeList.Add(dummyNode);

            TreeList.Node[] nodes = nodeList.ToArray();

            return nodes;
        }

        public List<SR1_Structure> GetProperties()
        {
            GoalTracker tracker = new GoalTracker();
            List<SR1_Structure> result = new List<SR1_Structure>();
            foreach (KeyValuePair<uint, SR1_Structure> entry in _Dictionary)
            {
                uint lookUpID = entry.Key;
                SR1_Structure structure = entry.Value;
                structure.GetProperties(tracker);

                string structTypeName = structure.GetTypeName(true);
                bool addStruct = false;
                if (structTypeName.StartsWith("StreamUnitPortal"))
                {
                    addStruct = true;
                }
                if (structTypeName == "intro")
                {
                    addStruct = true;
                }
                if (structTypeName.StartsWith("Intro"))
                {
                    addStruct = true;
                }
                if (structTypeName.ToLower().Contains("light"))
                {
                    addStruct = true;
                }
                if (structTypeName.ToLower().Contains("version"))
                {
                    addStruct = true;
                }
                if (addStruct)
                {
                    result.Add(structure);
                }
                //if (tracker.GoalAchieved)
                //{
                //    break;
                //}
            }

            return result;
        }
    }
}
