﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Recombobulator
{
    class SR1_PointerArray<T> : SR1_Structure where T : SR1_Structure, new()
    {
        private SR1_Pointer<T>[] _array = null;

        public SR1_PointerArray(int arrayLength, bool shouldCreate)
        {
            if (arrayLength > 0)
            {
                _array = new SR1_Pointer<T>[arrayLength];
                for (int i = 0; i < arrayLength; i++)
                {
                    _array[i] = new SR1_Pointer<T>(shouldCreate);
                }
            }
        }

        public override TreeList.Node[] CreatePropertyNodes()
        {
            if (_array.Length > 0)
            {
                TreeList.Node[] elementNodes = new TreeList.Node[_array.Length];
                for (int i = 0; i < _array.Length; i++)
                {
                    string elementTypeName = GetTypeName(false);
                    string elementName = "[" + i.ToString() + "]";
                    string elementValue = "0x" + _array[i].Offset.ToString("X8");

                    elementNodes[i] = new TreeList.Node(new string[] { "", elementTypeName, elementName, elementValue });
                }

                return elementNodes;
            }

            return null;
        }

        public override void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            if (_array != null)
            {
                for (int i = 0; i < _array.Length; i++)
                {
                    _array[i].Offset = reader.ReadUInt32();
                }
            }

            End = (uint)reader.BaseStream.Position;

            if (_array != null)
            {
                for (int i = 0; i < _array.Length; i++)
                {
                    try
                    {
                        if (_array[i].Offset != 0x00000000 && !dictionary.ContainsKey(_array[i].Offset))
                        {
                            reader.BaseStream.Position = (long)_array[i].Offset;
                            SR1_Structure structure = (SR1_Structure)_array[i].CreateObject(this, reader, tracker);
                            structure.Read(dictionary, reader, false, tracker);
                        }
                    }
                    catch (Exception e)
                    {

                    }
                }
            }
        }

        public override string ToString()
        {
            if ((_array is null) || (_array.Length == 0))
            {
                return "{ }";
            }

            return "{...}";
        }

        public override string GetTypeName(bool includeDimensions)
        {
            string typeName = typeof(T).Name + "*";

            if (_array == null)
            {
                typeName += "[]";
                return typeName;
            }

            if (includeDimensions)
            {
                typeName += "[";
                typeName += _array.Length.ToString();
                typeName += "]";
            }

            return typeName;
        }
    }
}
