﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Recombobulator
{
    class SR1_PrimitiveArray<T> : SR1_Structure// where T : struct
    {
        private int[] _dimensions = null;

        private T[] _array = null;

        protected const int _MaxArrayLengthToString = 64;
        protected const int _MaxArrayLength = 128;

        public SR1_PrimitiveArray(params int[] dimensions)
        {
            _dimensions = new int[dimensions.Length];
            dimensions.CopyTo(_dimensions, 0);

            int arrayLength = 1;
            foreach (int d in dimensions)
            {
                arrayLength *= d;
            }

            if (arrayLength > 0)
            {
                _array = new T[arrayLength];
            }
        }

        public override List<Property> GetProperties(GoalTracker tracker)
        {
            Properties = new List<Property>();

            if (_array == null)
            {
                return null;
            }

            for (int i = 0; i < _array.Length; i++)
            {
                Property newProp = new Property();
                newProp.TypeName = GetTypeName(false);
                newProp.PropertyName = "";
                newProp.PropertyValue = _array[i].ToString();

                int index = i;
                int d = _dimensions.Length;
                while (d > 0)
                {
                    d--;
                    int subIndex = index % _dimensions[d];
                    index /= _dimensions[d];
                    string indexName = "";
                    indexName += "[";
                    indexName += subIndex.ToString();
                    indexName += "]";
                    indexName += newProp.PropertyName;
                    newProp.PropertyName = indexName;
                }

                Properties.Add(newProp);
                if (i > _MaxArrayLength)
                {
                    break;
                }
            }

            return Properties;
        }

        public override TreeList.Node[] CreatePropertyNodes()
        {
            if (_array == null)
            {
                return null;
            }

            TreeList.Node[] elementNodes = new TreeList.Node[_array.Length];
            for (int i = 0; i < _array.Length; i++)
            {
                string typeName = GetTypeName(false);
                string elementName = "";
                string elementValue = _array[i].ToString();

                int index = i;
                int d = _dimensions.Length;
                while (d > 0)
                {
                    d--;
                    int subIndex = index % _dimensions[d];
                    index /= _dimensions[d];
                    string indexName = "";
                    indexName += "[";
                    indexName += subIndex.ToString();
                    indexName += "]";
                    indexName += elementName;
                    elementName = indexName;
                }

                elementNodes[i] = new TreeList.Node(new string[] { "", typeName, elementName, elementValue });
            }

            return elementNodes;
        }

        public override void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            if (_array != null)
            {
                for (int i = 0; i < _array.Length; i++)
                {
                    _array[i] = (T)ReadPrimitiveType(reader, Type.GetTypeCode(typeof(T)));
                }
            }

            End = (uint)reader.BaseStream.Position;
        }

        public override string ToString()
        {
            int stringLength = 0;
            if (typeof(T) == typeof(char))
            {
                if (_array != null && _dimensions.Length == 1)
                {
                    string result = "";
                    foreach (object o in _array)
                    {
                        if ((char)o == '\0')
                        {
                            break;
                        }

                        result += (char)o;
                        stringLength++;
                        if (stringLength > _MaxArrayLengthToString)
                        {
                            break;
                        }
                    }
                    return result;
                }

                return "";
            }
            else
            {
                if (_array == null)
                {
                    return "{ }";
                }

                if (_dimensions.Length == 1)
                {
                    string result = "{ ";
                    foreach (object o in _array)
                    {
                        string oString = o.ToString();
                        stringLength += oString.Length;
                        if (stringLength > _MaxArrayLengthToString)
                        {
                            break;
                        }

                        result += oString;
                        result += ", ";
                        
                    }
                    result = result.Trim(',', ' ');
                    result += " }";
                    return result;
                }

                return "{...}";
            }
        }

        public override string GetTypeName(bool includeDimensions)
        {
            string typeName = GetPrimitiveTypeName(Type.GetTypeCode(typeof(T)));

            if (_array == null)
            {
                typeName += "[]";
                return typeName;
            }

            if (includeDimensions)
            {
                int rank = _dimensions.Length;
                for (int r = 0; r < rank; r++)
                {
                    typeName += "[";
                    typeName += _dimensions[r].ToString();
                    typeName += "]";
                }
            }

            return typeName;
        }
    }
}
