﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace Recombobulator
{
    public class Property
    {
        public string TypeName;
        public string PropertyName;
        public string PropertyValue;

        public List<Property> Children;

        public Property()
        {
            Children = new List<Property>();
        }
    }

    class SR1_Structure
    {
        [SR1_Hidden]
        public uint Start { get; protected set; }
        [SR1_Hidden]
        public uint End { get; protected set; }
        [SR1_Hidden]
        public bool IsMember { get; protected set; }

        public List<Property> Properties;

        public SR1_Structure()
        {
            Properties = new List<Property>();
        }

        public virtual List<Property> GetProperties(GoalTracker tracker)
        {
            Type type = GetType();

            FieldInfo[] properties = type.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
            Properties = new List<Property>();

            for (int i = 0; i < properties.Length; i++)
            {
                //if (SR1_Hidden.IsHidden(properties[i]))
                //{
                //    continue;
                //}
                //if (tracker.GoalAchieved)
                //{
                //    break;
                //}

                Type fieldType = properties[i].FieldType;
                Property newProp = new Property();

                if (fieldType.IsGenericType &&
                    (fieldType.GetGenericTypeDefinition() == typeof(SR1_Pointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PrimitiveArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureSeriesPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PointerArrayPointer<>)))
                {
                    // don't need these for purposes of extracting the portal data
                    //continue;
                    Type genericFieldType = fieldType.GetGenericArguments()[0];
                    if (genericFieldType.IsPrimitive)
                    {
                        newProp.TypeName = GetPrimitiveTypeName(Type.GetTypeCode(genericFieldType));
                    }
                    else if (genericFieldType.IsClass && genericFieldType.BaseType == null)
                    {
                        newProp.TypeName = "void";
                    }
                    else
                    {
                        newProp.TypeName = genericFieldType.Name;
                    }
                    newProp.TypeName += "*";

                    newProp.PropertyName = properties[i].Name;

                    ISR1_Pointer pointer = (ISR1_Pointer)properties[i].GetValue(this);
                    newProp.PropertyValue = "0x" + pointer.Offset.ToString("X8");
                }
                else if (fieldType.BaseType == typeof(SR1_Structure))
                {
                    SR1_Structure structure = (SR1_Structure)(properties[i].GetValue(this));
                    //if (!tracker.GoalAchieved)
                    //{
                        string structTypeName = structure.GetTypeName(true);
                        //bool continueProcessing = false;
                        bool continueProcessing = true;
                        if (structTypeName.StartsWith("StreamUnitPortal"))
                        {
                            continueProcessing = true;
                        }
                        if (structTypeName == "intro")
                        {
                            continueProcessing = true;
                        }
                        if (structTypeName.StartsWith("Intro"))
                        {
                            continueProcessing = true;
                        }
                        if (structTypeName.ToLower().Contains("light"))
                        {
                            continueProcessing = true;
                        }
                        if (structTypeName.ToLower().Contains("version"))
                        {
                            continueProcessing = true;
                        }
                    //if (structTypeName.StartsWith("char"))
                    //{
                    //    continueProcessing = true;
                    //}
                    if (continueProcessing)
                        {
                            newProp.Children = structure.GetProperties(tracker);
                        }
                    //}

                    newProp.TypeName = structure.GetTypeName(true);
                    newProp.PropertyName = properties[i].Name;
                    newProp.PropertyValue = structure.ToString();
                }
                else if (fieldType.IsPrimitive == true)
                {
                    newProp.TypeName = GetPrimitiveTypeName(Type.GetTypeCode(fieldType));
                    newProp.PropertyName = properties[i].Name;
                    newProp.PropertyValue = properties[i].GetValue(this).ToString();
                }

                //if ((newProp != null) && (newProp.TypeName != null) && (newProp.TypeName.StartsWith("StreamUnitPortal[")))
                //{
                //    tracker.SetAchieved();
                //    //Properties.Add(newProp);
                //}

                Properties.Add(newProp);
            }

            return Properties;
        }

        public virtual TreeList.Node[] CreatePropertyNodes()
        {
            Type type = GetType();

            //PropertyInfo[] properties = GetType().GetProperties();
            FieldInfo[] properties = type.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
            TreeList.Node[] propertyNodes = new TreeList.Node[properties.Length];
            for (int i = 0; i < properties.Length; i++)
            {
                //if (SR1_Hidden.IsHidden(properties[i]))
                //{
                //    continue;
                //}

                Type fieldType = properties[i].FieldType;
                TreeList.Node[] childNodes = null;

                string typeName = "";
                string propertyName = "";
                string propertyValue = "";

                if (fieldType.IsGenericType &&
                    (fieldType.GetGenericTypeDefinition() == typeof(SR1_Pointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PrimitiveArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureSeriesPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PointerArrayPointer<>)))
                {
                    Type genericFieldType = fieldType.GetGenericArguments()[0];
                    if (genericFieldType.IsPrimitive)
                    {
                        typeName = GetPrimitiveTypeName(Type.GetTypeCode(genericFieldType));
                    }
                    else if (genericFieldType.IsClass && genericFieldType.BaseType == null)
                    {
                        typeName = "void";
                    }
                    else
                    {
                        typeName = genericFieldType.Name;
                    }
                    typeName += "*";

                    propertyName = properties[i].Name;

                    ISR1_Pointer pointer = (ISR1_Pointer)properties[i].GetValue(this);
                    propertyValue = "0x" + pointer.Offset.ToString("X8");
                }
                else if (fieldType.BaseType == typeof(SR1_Structure))
                {
                    SR1_Structure structure = (SR1_Structure)(properties[i].GetValue(this));

                    typeName = structure.GetTypeName(true);
                    propertyName = properties[i].Name;
                    propertyValue = structure.ToString();

                    childNodes = structure.CreatePropertyNodes();
                }
                else if (fieldType.IsPrimitive == true)
                {
                    typeName = GetPrimitiveTypeName(Type.GetTypeCode(fieldType));
                    propertyName = properties[i].Name;
                    propertyValue = properties[i].GetValue(this).ToString();
                }

                propertyNodes[i] = new TreeList.Node(new string[] { "", typeName, propertyName, propertyValue });

                if (childNodes != null)
                {
                    propertyNodes[i].Nodes.AddRange(childNodes);
                }
            }

            return propertyNodes;
        }

        public virtual void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            Type type = GetType();

            //PropertyInfo[] properties = GetType().GetProperties();
            FieldInfo[] properties = type.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
            TreeList.Node[] propertyNodes = new TreeList.Node[properties.Length];
            for (int i = 0; i < properties.Length; i++)
            {
                //if (SR1_Hidden.IsHidden(properties[i]))
                //{
                //    continue;
                //}

                Type fieldType = properties[i].FieldType;

                if (fieldType.IsGenericType &&
                    (fieldType.GetGenericTypeDefinition() == typeof(SR1_Pointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PrimitiveArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureSeriesPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PointerArrayPointer<>)))
                {
                    ISR1_Pointer pointer = (ISR1_Pointer)properties[i].GetValue(this);
                    pointer.Offset = reader.ReadUInt32();
                }
                else if (fieldType.BaseType == typeof(SR1_Structure))
                {
                    //if (!tracker.GoalAchieved)
                    //{
                        SR1_Structure structure = (SR1_Structure)properties[i].GetValue(this);
                        structure.Read(dictionary, reader, true, tracker);
                    //}
                }
                else if (fieldType.IsPrimitive)
                {
                    properties[i].SetValue(this, ReadPrimitiveType(reader, Type.GetTypeCode(fieldType)));
                }
            }

            End = (uint)reader.BaseStream.Position;

            for (int i = 0; i < properties.Length; i++)
            {
                //if (SR1_Hidden.IsHidden(properties[i]))
                //{
                //    continue;
                //}
                //if (tracker.GoalAchieved)
                //{
                //    continue;
                //}

                Type fieldType = properties[i].FieldType;

                if (fieldType.IsGenericType &&
                    (fieldType.GetGenericTypeDefinition() == typeof(SR1_Pointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PrimitiveArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureArrayPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_StructureSeriesPointer<>) ||
                    fieldType.GetGenericTypeDefinition() == typeof(SR1_PointerArrayPointer<>)))
                {
                    try
                    {
                        ISR1_Pointer pointer = (ISR1_Pointer)properties[i].GetValue(this);
                        
                        if (pointer.Offset != 0x00000000 && !dictionary.ContainsKey(pointer.Offset))
                        {
                            reader.BaseStream.Position = (long)pointer.Offset;
                            SR1_Structure structure = (SR1_Structure)pointer.CreateObject(this, reader, tracker);
                            structure.Read(dictionary, reader, false, tracker);
                        }

                        //if (pointer.GetType() == typeof(SR1_Pointer<Recombobulator.SR1Structures.StreamUnitPortalList>))
                        //{
                        //    //tracker.SetAchieved();
                        //    Console.WriteLine("Found Portal List");
                        //}
                    }
                    catch (Exception e)
                    {

                    }
                }
                //if (tracker.GoalAchieved)
                //{
                //    break;
                //}
            }

            reader.BaseStream.Position = End;
        }

        protected object ReadPrimitiveType(BinaryReader reader, TypeCode typeCode)
        {
            switch (typeCode)
            {
                case TypeCode.Boolean:
                    return reader.ReadBoolean();
                case TypeCode.Char:
                    return (char)reader.ReadSByte();
                case TypeCode.SByte:
                    return reader.ReadSByte();
                case TypeCode.Byte:
                    return reader.ReadByte();
                case TypeCode.Int16:
                    return reader.ReadInt16();
                case TypeCode.UInt16:
                    return reader.ReadUInt16();
                case TypeCode.Int32:
                    return reader.ReadInt32();
                case TypeCode.UInt32:
                    return reader.ReadUInt32();
                case TypeCode.Int64:
                    return reader.ReadInt64();
                case TypeCode.UInt64:
                    return reader.ReadUInt64();
                case TypeCode.Single:
                    return reader.ReadSingle();
                case TypeCode.Double:
                    return reader.ReadDouble();
                default:
                    throw new Exception("Unhandled primitive type.");
            }
        }

        public override string ToString()
        {
            return "";
        }

        public virtual string GetTypeName(bool includeDimensions)
        {
            return GetType().Name;
        }

        protected string GetPrimitiveTypeName(TypeCode typeCode)
        {
            switch (typeCode)
            {
                case TypeCode.Boolean:
                    return "bool";
                case TypeCode.Char:
                case TypeCode.SByte:
                    return "char";
                case TypeCode.Byte:
                    return "unsigned char";
                case TypeCode.Int16:
                    return "short";
                case TypeCode.UInt16:
                    return "unsigned short";
                case TypeCode.Int32:
                    return "long"; // "int";
                case TypeCode.UInt32:
                    return "unsigned int";
                case TypeCode.Int64:
                    return "long long";
                case TypeCode.UInt64:
                    return "unsigned long long";
                case TypeCode.Single:
                    return "float";
                case TypeCode.Double:
                    return "double";
                default:
                    throw new Exception("Unhandled primitive type.");
            }
        }
    };
}
