﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Recombobulator
{
    class SR1_StructureSeries<T> : SR1_Structure where T : SR1_Structure, new()
    {
        private int _BufferLength;

        private List<T> _List = new List<T>();

        protected const int _MaxArrayLength = 128;

        public SR1_StructureSeries(int bufferLength)
        {
            _BufferLength = bufferLength;
        }

        public override List<Property> GetProperties(GoalTracker tracker)
        {
            Properties = new List<Property>();

            if (_List.Count == 0)
            {
                return null;
            }

            for (int i = 0; i < _List.Count; i++)
            {
                Property newProp = new Property();
                newProp.TypeName = GetTypeName(false);
                newProp.PropertyName = "[" + i.ToString() + "]";
                newProp.PropertyValue = _List[i].ToString();

                //if (!tracker.GoalAchieved)
                //{
                    newProp.Children = _List[i].GetProperties(tracker);
                //}
                Properties.Add(newProp);
                if (i > _MaxArrayLength)
                {
                    break;
                }
            }

            return Properties;
        }

        public override TreeList.Node[] CreatePropertyNodes()
        {
            if (_List.Count == 0)
            {
                return null;
            }

            TreeList.Node[] elementNodes = new TreeList.Node[_List.Count];
            for (int i = 0; i < _List.Count; i++)
            {
                string typeName = GetTypeName(false);
                string elementName = "[" + i.ToString() + "]";
                string elementValue = _List[i].ToString();

                elementNodes[i] = new TreeList.Node(new string[] { "", typeName, elementName, elementValue });
                elementNodes[i].Nodes.AddRange(_List[i].CreatePropertyNodes());
            }

            return elementNodes;
        }

        public override void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            long endPosition = reader.BaseStream.Position + _BufferLength;
            while (reader.BaseStream.Position < endPosition)
            {
                T newEntry = new T();
                newEntry.Read(dictionary, reader, true, tracker);
                _List.Add(newEntry);
            }

            End = (uint)reader.BaseStream.Position;
        }

        public override string ToString()
        {
            if (_List.Count == 0)
            {
                return "{ }";
            }

            return "{...}";
        }

        public override string GetTypeName(bool includeDimensions)
        {
            string typeName = typeof(T).Name;

            if (_List.Count == 0)
            {
                typeName += "[]";
                return typeName;
            }

            if (includeDimensions)
            {
                typeName += "[";
                typeName += _List.Count;
                typeName += "]";
            }

            return typeName;
        }
    }
}
