#!/usr/bin/env python3

import bmesh
import bpy
import math
import re
import time
from operator import attrgetter

gWriteToLogFile = True
gLineSeparator = "\n"
highlightColour = (0.0, 1.0, 0.0, 1.0)
#highlightColourDeletedArea = (0.0, 0.0, 1.0, 1.0)
highlightColourDeletedArea = (0.004, 0.152, 1.0, 1.0)
outlineColour = (0.0, 0.0, 0.0, 1.0)
boxAlpha = 0.1
outlineAlpha = 1.0
textAlpha = 1.0
textFont = bpy.data.fonts.load('C:\\Windows\\Fonts\\FiraSans-Bold.otf')
#textOffsets= [(10, 10), (12.5, 15), (15, 10)]
textOffsets= [(20, 10), (22.5, 7.5), (25, 5)]
textOffsetNumNorthWest = 0
textOffsetNumNorthEast = 0
textOffsetNumSouthWest = 0
textOffsetNumSouthEast = 0
#20.0, textOffsetY = 10

def WriteToFile(outputFilePath, content):
	global gWriteToLogFile
	if gWriteToLogFile and outputFilePath != "":
		try:
			outputFile = open(outputFilePath, "a", newline=gLineSeparator)
			outputFile.write(content)
			outputFile.write(gLineSeparator)
			outputFile.close()
		except Exception as e:
			gWriteToLogFile = False
			print("Error writing to file '%s': %s. Logging will be disabled." % (outputFilePath, e))
	else:
		print(content)

def getNodeByName(material, nodeName):
	for n in material.node_tree.nodes:
		if n.name == nodeName:
			return n
	return None

class BlenderObject:

	def ResetBoundingBox(self):
		initialMin = 10000.0
		initialMax = -10000.0
		self.MinX = initialMin
		self.MaxX = initialMax
		self.MinY = initialMin
		self.MaxY = initialMax
		self.MinZ = initialMin
		self.MaxZ = initialMax
		self.AbsoluteMinX = abs(initialMin)
		self.AbsoluteMaxX = 0
		self.AbsoluteMinY = abs(initialMin)
		self.AbsoluteMaxY = 0
		self.AbsoluteMinZ = abs(initialMin)
		self.AbsoluteMaxZ = 0
		self.MinEuclideanDistanceFromOrigin = 0
		self.MaxEuclideanDistanceFromOrigin = 0

	def __init__(self):
		self.ResetBoundingBox()
		self.Name = ""
		self.ScriptPlacementOrder = 100000
		self.DistanceFromOrigin = 0
		self.G2Type = "unknown"
		self.BaseBlenderObject = None

	def GetSizeX(self):
		return (self.MaxX - self.MinX)

	def GetMidX(self):
		return (self.GetSizeX() / 2.0) + self.MinX

	def GetSizeY(self):
		return (self.MaxY - self.MinY)

	def GetMidY(self):
		return (self.GetSizeY() / 2.0) + self.MinY

	def GetSizeZ(self):
		return (self.MaxZ - self.MinZ)

	def GetMidZ(self):
		return (self.GetSizeZ() / 2.0) + self.MinZ
		
	def Clone(self):
		result = BlenderObject()
		result.MinX = self.MinX
		result.MaxX = self.MaxX
		result.MinY = self.MinY
		result.MaxY = self.MaxY
		result.MinZ = self.MinZ
		result.MaxZ = self.MaxZ
		result.ScriptPlacementOrder = self.ScriptPlacementOrder
		result.DistanceFromOrigin = self.DistanceFromOrigin
		result.G2Type = self.G2Type
		result.SetDerivedValues()
		return result
	
	def SetDerivedValues(self):
		#self.AbsoluteMinX = abs(self.MinX)
		#self.AbsoluteMaxX = abs(self.MaxX)
		#self.AbsoluteMinY = abs(self.MinY)
		#self.AbsoluteMaxY = abs(self.MaxY)
		#self.AbsoluteMinZ = abs(self.MinZ)
		#self.AbsoluteMaxZ = abs(self.MaxZ)
		self.MinEuclideanDistanceFromOrigin = math.sqrt(pow(self.AbsoluteMinX, 2) + pow(self.AbsoluteMinY, 2) + pow(self.AbsoluteMinZ, 2))
		self.MaxEuclideanDistanceFromOrigin = math.sqrt(pow(self.AbsoluteMaxX, 2) + pow(self.AbsoluteMaxY, 2) + pow(self.AbsoluteMaxZ, 2))
		
	def SetVertexBoundingBox(self, childObject):
		verts = [(childObject.matrix_world @ v.co) for v in childObject.data.vertices]
		for vert in verts:
			if vert[0] < self.MinX:
				self.MinX = vert[0]
			if vert[0] > self.MaxX:
				self.MaxX = vert[0]

			if abs(vert[0]) < self.AbsoluteMinX:
				self.AbsoluteMinX = abs(vert[0])
			if abs(vert[0]) > self.AbsoluteMaxX:
				self.AbsoluteMaxX = abs(vert[0])
				
			if vert[1] < self.MinY:
				self.MinY = vert[1]
			if vert[1] > self.MaxY:
				self.MaxY = vert[1]
				
			if abs(vert[1]) < self.AbsoluteMinY:
				self.AbsoluteMinY = abs(vert[1])
			if abs(vert[1]) > self.AbsoluteMaxY:
				self.AbsoluteMaxY = abs(vert[1])
				
			if vert[2] < self.MinZ:
				self.MinZ = vert[2]
			if vert[2] > self.MaxZ:
				self.MaxZ = vert[2]

			if abs(vert[2]) < self.AbsoluteMinZ:
				self.AbsoluteMinZ = abs(vert[2])
			if abs(vert[2]) > self.AbsoluteMaxZ:
				self.AbsoluteMaxZ = abs(vert[2])
				
		self.SetDerivedValues()
	
	def SetVertexBoundingBoxUsingTextCurve(self, textCurveObject):
		bpy.context.view_layer.update()
		tcHalfWidth = (textCurveObject.dimensions[0] / 2.0)
		tcMinX = textCurveObject.location[0] - tcHalfWidth
		tcMaxX = textCurveObject.location[0] + tcHalfWidth

		tcHalfDepth = (textCurveObject.dimensions[1] / 2.0)
		tcMinY = textCurveObject.location[1] - tcHalfDepth
		tcMaxY = textCurveObject.location[1] + tcHalfDepth

		tcHalfHeight = (textCurveObject.dimensions[2] / 2.0)
		tcMinZ = textCurveObject.location[2] - tcHalfHeight
		tcMaxZ = textCurveObject.location[2] + tcHalfHeight
		
		if tcMinX < self.MinX:
			self.MinX = tcMinX
		if tcMaxX > self.MaxX:
			self.MaxX = tcMaxX
		if tcMinY < self.MinY:
			self.MinY = tcMinY
		if tcMaxY > self.MaxY:
			self.MaxY = tcMaxY
		if tcMinZ < self.MinZ:
			self.MinZ = tcMinZ
		if tcMaxZ > self.MaxZ:
			self.MaxZ = tcMaxZ
			
		tcAbsoluteMinX = abs(tcMinX)
		tcAbsoluteMaxX = abs(tcMaxX)
		tcAbsoluteMinY = abs(tcMinY)
		tcAbsoluteMaxY = abs(tcMaxY)
		tcAbsoluteMinZ = abs(tcMinZ)
		tcAbsoluteMaxZ = abs(tcMaxZ)
		
		if tcAbsoluteMinX < self.AbsoluteMinX:
			self.AbsoluteMinX = tcAbsoluteMinX
		if tcAbsoluteMaxX > self.AbsoluteMaxX:
			self.AbsoluteMaxX = tcAbsoluteMaxX
		if tcAbsoluteMinY < self.AbsoluteMinY:
			self.AbsoluteMinY = tcAbsoluteMinY
		if tcAbsoluteMaxY > self.AbsoluteMaxY:
			self.AbsoluteMaxY = tcAbsoluteMaxY
		if tcAbsoluteMinZ < self.AbsoluteMinZ:
			self.AbsoluteMinZ = tcAbsoluteMinZ
		if tcAbsoluteMaxZ > self.AbsoluteMaxZ:
			self.AbsoluteMaxZ = tcAbsoluteMaxZ
		
		self.SetDerivedValues()

	def SetVertexBoundingBoxFromOtherObject(self, otherObject):
		self.MinX = otherObject.MinX
		self.MaxX = otherObject.MaxX
		self.MinY = otherObject.MinY
		self.MaxY = otherObject.MaxY
		self.MinZ = otherObject.MinZ
		self.MaxZ = otherObject.MaxZ

		self.AbsoluteMinX = otherObject.AbsoluteMinX
		self.AbsoluteMaxX = otherObject.AbsoluteMaxX
		self.AbsoluteMinY = otherObject.AbsoluteMinY
		self.AbsoluteMaxY = otherObject.AbsoluteMaxY
		self.AbsoluteMinZ = otherObject.AbsoluteMinZ
		self.AbsoluteMaxZ = otherObject.AbsoluteMaxZ
		
		self.SetDerivedValues()

# https://blenderartists.org/t/how-to-get-a-list-of-an-objects-children/465508
def getChildren(myObject): 
    children = [] 
    for ob in bpy.data.objects: 
        if ob.parent == myObject: 
            children.append(ob) 
    return children 

def getResultByName(resultList, name):
	for r in resultList:
		if r.Name == name:
			return r
	return None

def getNewCube(castShadows = False):
    existing_objects = set(bpy.context.scene.objects)
    bpy.ops.mesh.primitive_cube_add(size=1, enter_editmode=False, location=(0, 0, 0))
    new_cube_collection = set(bpy.context.scene.objects) - existing_objects
    result = None
    for nc in new_cube_collection:
		if not castShadows:
			nc.cycles_visibility.shadow = False
        result = nc
    return result

def getSortableObjectListInner(existingResults, g2type, debugLogPath = "", areaNameWhitelistRegex = None):
	for o in bpy.data.objects:
		#if o and o.data:
		includeObject = True
		#if "." in o.name or "-" in o.name:
		#	includeObject = False
		if o.parent and o.parent != "":
			includeObject = False
		if "g2_type" in o.keys():
			if o['g2_type'] != g2type:
				includeObject = False
		if includeObject:
			newObject = BlenderObject()
			newObject.Name = o.name
			newObject.BaseBlenderObject = o
			#WriteToFile(debugLogPath, o.name)
			#WriteToFile(debugLogPath, o.keys())
			if "scripted_placement_order" in o.keys():
				newObject.ScriptPlacementOrder = int(o['scripted_placement_order'])
			else:
				WriteToFile(debugLogPath, "Warning: did not find a placement order property for object '%s'" % (o.name))
				includeObject = False
			if "distance_from_origin" in o.keys():
				newObject.DistanceFromOrigin = int(o['distance_from_origin'])
			else:
				WriteToFile(debugLogPath, "Warning: did not find a distance from origin property for object '%s'" % (o.name))
				includeObject = False
			if "g2_type" in o.keys():
				newObject.G2Type = o['g2_type']
			else:
				WriteToFile(debugLogPath, "Warning: did not find a G2 element type property for object '%s'" % (o.name))
				includeObject = False
			if newObject.G2Type == "area":
				areaNameTemp = newObject.Name
				if areaNameTemp.startswith("area_"):
					areaNameTemp = areaNameTemp[5:]
				if areaNameWhitelistRegex is not None and not re.match(areaNameWhitelistRegex, areaNameTemp):
					WriteToFile(debugLogPath, "Warning: ignoring area '%s', because it does not match the whitelist regular expression '%s'" % (o.name, areaNameWhitelistRegex))
					includeObject = False
			
			if includeObject:
				children = getChildren(o)
				# for objects, use the position data from the parent area, so that they're arranged in a more pleasing order
				usedParentPosition = False
				if newObject.G2Type == "object":
					parentAreaName = "area_%s" % (newObject.Name.split('_')[0])
					parentArea = getResultByName(existingResults, parentAreaName)
					if parentArea:
						newObject.MinX = float(parentArea.MinX)
						newObject.MaxX = float(parentArea.MaxX)
						newObject.MinY = float(parentArea.MinY)
						newObject.MaxY = float(parentArea.MaxY)
						newObject.MinZ = float(parentArea.MinZ)
						newObject.MaxZ = float(parentArea.MaxZ)
						newObject.SetDerivedValues()
						usedParentPosition = True
					else:
						includeObject = False
						WriteToFile(debugLogPath, "Did not find a parent area for object '%s' - this object will be ignored" % (o.name))
				if not usedParentPosition:
					for c in children: 
						if not c.data:
							WriteToFile(debugLogPath, "No data for child object '%s'" % (o.name))
							#includeObject = False
						else:
							#if newObject.G2Type == "object":
							#	WriteToFile(debugLogPath, "No parent for G2 object '%s' - using its own coordinates" % (o.name))
							newObject.SetVertexBoundingBox(c)
							# verts = [(c.matrix_world @ v.co) for v in c.data.vertices]
							# for vert in verts:
								# if vert[0] < newObject.MinX:
									# newObject.MinX = vert[0]
								# if vert[0] > newObject.MaxX:
									# newObject.MaxX = vert[0]
								# if vert[1] < newObject.MinY:
									# newObject.MinY = vert[1]
								# if vert[1] > newObject.MaxY:
									# newObject.MaxY = vert[1]
								# if vert[2] < newObject.MinZ:
									# newObject.MinZ = vert[2]
								# if vert[2] > newObject.MaxZ:
									# newObject.MaxZ = vert[2]
		if includeObject:
			existingResults.append(newObject)
	return existingResults
	
def getSortableObjectList(debugLogPath = "", areaNameWhitelistRegex = None):
	#obj = bpy.data.objects['Cube']
	result = []
	# this is done twice, first for areas, then objects
	# because otherwise if the area name falls before "area_" in the alphabet, the area won't be present in the 
	# results to bind child objects to, and they won't be bound
	result = getSortableObjectListInner(result, "area", debugLogPath = debugLogPath, areaNameWhitelistRegex = areaNameWhitelistRegex)
	result = getSortableObjectListInner(result, "object", debugLogPath = debugLogPath, areaNameWhitelistRegex = areaNameWhitelistRegex)
	return result

def getAreaWithObjectsContainer(areaName, debugLogPath = ""):
	awo_name = "awo_%s" % (areaName)
	awo = None
	if awo_name in bpy.data.collections.keys():
		WriteToFile(debugLogPath, 'Using existing area-with-objects %s' % (awo_name))
		awo = bpy.data.collections[awo_name]
	else:
		WriteToFile(debugLogPath, 'Creating area-with-objects %s' % (awo_name))
		WriteToFile(debugLogPath, 'Creating area-with-objects %s' % (awo_name))
		awo = bpy.data.collections.new(name=awo_name)	
	return awo

def drawTextCenteredAtLocation(g2Object, bodyText, textFont, fontSize, baseName, position, textColour):
	newEntries = []
	font_curve1 = bpy.data.curves.new(type  ="FONT", name = "bounding_box_font_curve1_%s" % (baseName))
	font_object1 = bpy.data.objects.new("bounding_box_font_object1_%s" % (baseName), font_curve1)
	font_object1.data.body = bodyText
	font_object1.data.size = fontSize
	font_object1.data.offset = 0.4
	font_object1.data.fill_mode = 'BOTH'
	font_object1.data.align_x = 'CENTER'
	font_object1.data.align_y = 'CENTER'
	font_object1.data.font = textFont
	font_object1.data.extrude = 0.1
	font_object1.location = (position[0], position[1], (position[2] + 0.3))
	font_object1.cycles_visibility.shadow = False
	matf1 = bpy.data.materials.new(name = 'm_bounding_box_font_object1_%s' % (baseName))
	matf1.use_nodes= True
	matf1.blend_method = 'BLEND'
	bsdfNodeF1 = None
	for n in matf1.node_tree.nodes:
		if n.name == 'Principled BSDF':
			bsdfNodeF1 = n
	if bsdfNodeF1:
		bsdfNodeF1.inputs['Base Color'].default_value = outlineColour
		bsdfNodeF1.inputs['Emission'].default_value = outlineColour
		bsdfNodeF1.inputs['Specular'].default_value = 0.0
		bsdfNodeF1.inputs['Alpha'].default_value = outlineAlpha
		outlineAlpha
	if font_object1.data.materials:
		font_object1.data.materials[0] = matf1
	else:
		font_object1.data.materials.append(matf1)
	
	font_curve2 = bpy.data.curves.new(type  ="FONT", name = "bounding_box_font_curve2_%s" % (baseName))
	font_object2 = bpy.data.objects.new("bounding_box_font_object2_%s" % (baseName), font_curve2)
	font_object2.data.body = bodyText
	font_object2.data.size = fontSize
	font_object2.data.fill_mode = 'BOTH'
	font_object2.data.align_x = 'CENTER'
	font_object2.data.align_y = 'CENTER'
	font_object2.data.font = textFont
	font_object2.data.extrude = 0.1
	font_object2.location = (position[0], position[1], (position[2] + 0.4))
	font_object2.cycles_visibility.shadow = False
	matf2 = bpy.data.materials.new(name = 'm_bounding_box_font_object2_%s' % (baseName))
	matf2.use_nodes= True
	matf2.blend_method = 'BLEND'
	bsdfNodeF2 = None
	for n in matf2.node_tree.nodes:
		if n.name == 'Principled BSDF':
			bsdfNodeF2 = n
	if bsdfNodeF2:
		bsdfNodeF2.inputs['Base Color'].default_value = textColour
		bsdfNodeF2.inputs['Emission'].default_value = textColour
		bsdfNodeF2.inputs['Specular'].default_value = 0.0
		bsdfNodeF2.inputs['Alpha'].default_value = textAlpha
	if font_object2.data.materials:
		font_object2.data.materials[0] = matf2
	else:
		font_object2.data.materials.append(matf2)
	
	neFont1 = g2Object.Clone()
	neFont1.BaseBlenderObject = font_object1
	neFont1.Name = font_object1.name
	neFont1.ResetBoundingBox()
	neFont1.SetVertexBoundingBoxUsingTextCurve(font_object1)
	newEntries.append(neFont1)
	
	neFont2 = g2Object.Clone()
	neFont2.BaseBlenderObject = font_object2
	neFont2.Name = font_object2.name
	neFont2.ResetBoundingBox()
	neFont2.SetVertexBoundingBoxUsingTextCurve(font_object2)
	newEntries.append(neFont2)
	
	return newEntries

def drawBoxFrameAroundObject(g2Object, frameWidth, fontSize, existingNewEntries, frameColour, debugLogPath = ""):
	fwMultiplier = 1.0
	newEntries = []
	frameColour = highlightColour
	for i in range(0, len(existingNewEntries)):
		newEntries.append(existingNewEntries[i])
	if g2Object.BaseBlenderObject:
		baseName = g2Object.Name
		if baseName.startswith("area_"):
			baseName = baseName[5:]
		if isDeletedArea(baseName):
			frameColour = highlightColourDeletedArea
		#bpy.context.evaluated_depsgraph_get().update()
		elements = []
		
		cubeTopBack = getNewCube()
		cubeTopBack.name = 'bb_%s_topback' % (baseName)
		cubeTopBack.dimensions = (g2Object.GetSizeX() + (frameWidth * fwMultiplier), frameWidth, frameWidth)
		cubeTopBack.location = (g2Object.GetMidX(), g2Object.MaxY, g2Object.MaxZ)
		#awo.objects.link(bpy.data.objects[cubeTopBack.name])
		elements.append(cubeTopBack)
		
		cubeTopLeft = getNewCube()
		cubeTopLeft.name = 'bb_%s_topleft' % (baseName)
		cubeTopLeft.dimensions = (frameWidth, g2Object.GetSizeY() + (frameWidth * fwMultiplier), frameWidth)
		cubeTopLeft.location = (g2Object.MinX, g2Object.GetMidY(), g2Object.MaxZ)
		#awo.objects.link(bpy.data.objects[cubeTopLeft.name])
		elements.append(cubeTopLeft)
		
		cubeTopRight = getNewCube()
		cubeTopRight.name = 'bb_%s_topright' % (baseName)
		cubeTopRight.dimensions = (frameWidth, g2Object.GetSizeY() + (frameWidth * fwMultiplier), frameWidth)
		cubeTopRight.location = (g2Object.MaxX, g2Object.GetMidY(), g2Object.MaxZ)
		#awo.objects.link(bpy.data.objects[cubeTopRight.name])
		elements.append(cubeTopRight)
		
		cubeTopFront = getNewCube()
		cubeTopFront.name = 'bb_%s_topfront' % (baseName)
		cubeTopFront.dimensions = (g2Object.GetSizeX() + (frameWidth * fwMultiplier), frameWidth, frameWidth)
		cubeTopFront.location = (g2Object.GetMidX(), g2Object.MinY, g2Object.MaxZ)
		#awo.objects.link(bpy.data.objects[cubeTopFront.name])
		elements.append(cubeTopFront)

		cubeVertFrontLeft = getNewCube()
		cubeVertFrontLeft.name = 'bb_%s_vertfrontleft' % (baseName)
		cubeVertFrontLeft.dimensions = (frameWidth, frameWidth, g2Object.GetSizeZ() + (frameWidth * fwMultiplier))
		cubeVertFrontLeft.location = (g2Object.MinX, g2Object.MinY, g2Object.GetMidZ())
		#awo.objects.link(bpy.data.objects[cubeVertFrontLeft.name])
		elements.append(cubeVertFrontLeft)
		
		cubeVertFrontRight = getNewCube()
		cubeVertFrontRight.name = 'bb_%s_vertfrontright' % (baseName)
		cubeVertFrontRight.dimensions = (frameWidth, frameWidth, g2Object.GetSizeZ() + (frameWidth * fwMultiplier))
		cubeVertFrontRight.location = (g2Object.MinX, g2Object.MaxY, g2Object.GetMidZ())
		#awo.objects.link(bpy.data.objects[cubeVertFrontRight.name])
		elements.append(cubeVertFrontRight)
		
		cubeVertBackLeft = getNewCube()
		cubeVertBackLeft.name = 'bb_%s_vertbackleft' % (baseName)
		cubeVertBackLeft.dimensions = (frameWidth, frameWidth, g2Object.GetSizeZ() + (frameWidth * fwMultiplier))
		cubeVertBackLeft.location = (g2Object.MaxX, g2Object.MinY, g2Object.GetMidZ())
		#awo.objects.link(bpy.data.objects[cubeVertBackLeft.name])
		elements.append(cubeVertBackLeft)
		
		cubeVertBackRight = getNewCube()
		cubeVertBackRight.name = 'bb_%s_vertbackright' % (baseName)
		cubeVertBackRight.dimensions = (frameWidth, frameWidth, g2Object.GetSizeZ() + (frameWidth * fwMultiplier))
		cubeVertBackRight.location = (g2Object.MaxX, g2Object.MaxY, g2Object.GetMidZ())
		#awo.objects.link(bpy.data.objects[cubeVertBackRight.name])
		elements.append(cubeVertBackRight)
		
		cubeBottomBack = getNewCube()
		cubeBottomBack.name = 'bb_%s_bottomback' % (baseName)
		cubeBottomBack.dimensions = (g2Object.GetSizeX() + (frameWidth * fwMultiplier), frameWidth, frameWidth)
		cubeBottomBack.location = (g2Object.GetMidX(), g2Object.MaxY, g2Object.MinZ)
		#awo.objects.link(bpy.data.objects[cubeBottomBack.name])
		elements.append(cubeBottomBack)
					
		cubeBottomLeft = getNewCube()
		cubeBottomLeft.name = 'bb_%s_bottomleft' % (baseName)
		cubeBottomLeft.dimensions = (frameWidth, g2Object.GetSizeY() + (frameWidth * fwMultiplier), frameWidth)
		cubeBottomLeft.location = (g2Object.MinX, g2Object.GetMidY(), g2Object.MinZ)
		#awo.objects.link(bpy.data.objects[cubeBottomLeft.name])
		elements.append(cubeBottomLeft)
		
		cubeBottomRight = getNewCube()
		cubeBottomRight.name = 'bb_%s_bottomright' % (baseName)
		cubeBottomRight.dimensions = (frameWidth, g2Object.GetSizeY() + (frameWidth * fwMultiplier), frameWidth)
		cubeBottomRight.location = (g2Object.MaxX, g2Object.GetMidY(), g2Object.MinZ)
		#awo.objects.link(bpy.data.objects[cubeBottomRight.name])
		elements.append(cubeBottomRight)
		
		cubeBottomFront = getNewCube()
		cubeBottomFront.name = 'bb_%s_bottomfront' % (baseName)
		cubeBottomFront.dimensions = (g2Object.GetSizeX() + (frameWidth * fwMultiplier), frameWidth, frameWidth)
		cubeBottomFront.location = (g2Object.GetMidX(), g2Object.MinY, g2Object.MinZ)
		#awo.objects.link(bpy.data.objects[cubeBottomFront.name])
		elements.append(cubeBottomFront)

		#WriteToFile(debugLogPath, 'Creating union of %s' % (elements))
		bpy.context.evaluated_depsgraph_get().update()
		#ctx = bpy.context.copy()

		# one of the objects to join
		#ctx['active_object'] = elements[0]

		#ctx['selected_objects'] = elements			

		#bpy.ops.object.join(ctx)
		
		#for i in range(0, len(elements)):
		#	bpy.context.scene.collection.objects.link(elements[i])
		#bpy.context.evaluated_depsgraph_get().update()
		
		bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
		bpy.ops.object.select_all(action='DESELECT')
		elements[0].select_set(True)
		bpy.context.view_layer.objects.active = elements[0]			
		
		for i in range(1, len(elements)):
			elements[i].select_set(True)
			bpy.ops.object.join()
			#modName = 'umod_%s_%s' % (baseName, i)
			#umod = elements[0].modifiers.new(name=modName, type='BOOLEAN')
			#umod.operation = 'UNION'
			#umod.object = elements[i]
			#bpy.ops.object.select_all(action='DESELECT')
			#elements[0].select_set(True)
			#bpy.context.view_layer.objects.active = elements[0]
			#bpy.ops.object.modifier_apply(apply_as='DATA', modifier='umod%s' % (i))
			#bpy.ops.object.modifier_apply(apply_as='DATA', modifier=modName)
			#bpy.ops.object.modifier_apply({"object": elements[0]}, apply_as='DATA', modifier=modName)
			#elements[0].data = elements[0].to_mesh(preserve_all_data_layers = True, depsgraph = bpy.context.evaluated_depsgraph_get()).copy()
			#bpy.data.objects.unlink(elements[i])
			#bpy.context.scene.objects.unlink(elements[i])
			#bpy.ops.object.select_all(action='DESELECT')
			#bpy.context.scene.collection.objects.unlink(elements[i])
			#bpy.ops.wm.redraw_timer(type='DRAW_WIN', iterations=1)
			#time.sleep(.010)
		
		#bpy.context.evaluated_depsgraph_get().update()
		
		elements[0].name='bounding_box_%s_data' % (baseName)
		mat = bpy.data.materials.new(name = 'm_bounding_box_%s' % (baseName))
		mat.use_nodes= True
		mat.blend_method = 'BLEND'
		bsdfNode = None
		for n in mat.node_tree.nodes:
			if n.name == 'Principled BSDF':
				bsdfNode = n
		if bsdfNode:
			bsdfNode.inputs['Base Color'].default_value = frameColour
			bsdfNode.inputs['Emission'].default_value = frameColour
			bsdfNode.inputs['Specular'].default_value = 0.0
			bsdfNode.inputs['Alpha'].default_value = boxAlpha
		if elements[0].data.materials:
			elements[0].data.materials[0] = mat
		else:
			elements[0].data.materials.append(mat)
		
		bpy.ops.object.select_all(action='DESELECT')
		
		awo = getAreaWithObjectsContainer(baseName)
		awo.objects.link(elements[0])
		
		neFrame = g2Object.Clone()
		neFrame.BaseBlenderObject = elements[0]
		neFrame.Name = elements[0].name
		newEntries.append(neFrame)
		
		textEntries = drawTextCenteredAtLocation(g2Object, baseName, textFont, fontSize, baseName, (g2Object.GetMidX(), g2Object.GetMidY(), g2Object.MaxZ), frameColour)
		for i in range(0, len(textEntries)):
			newEntries.append(textEntries[i])
			awo.objects.link(textEntries[i].BaseBlenderObject)
		
		#bpy.data.collections[awo.name].objects.link[elements[0]]
		#parentAreaWithObjects.objects.link(bpy.data.objects[objectName])
		
		#bpy.context.scene.objects.link(mesh_object)
	return newEntries

def addBoxesToObjectsInList(objectList, debugLogPath = ""):
	frameWidth = 0.5
	fontSize = 6.0
	newEntries = []

	for g2o in objectList:
		newEntries = drawBoxFrameAroundObject(g2o, frameWidth, fontSize, newEntries, debugLogPath = debugLogPath)

	for ne in newEntries:
		objectList.append(ne)

	return objectList


def addBracketFrameAroundObject(g2Object, baseName, frameWidth, bracketLength, debugLogPath = "", minBracketGap = 0.5, padding = 0.0):
	newEntries = []
	if g2Object.BaseBlenderObject:
		baseName = g2Object.Name
		if baseName.startswith("area_"):
			baseName = baseName[5:]
		#bpy.context.evaluated_depsgraph_get().update()
		elements = []
		bracketPosOffsetLongAxis = (bracketLength / 2.0)
		bracketPosOffsetShortAxis = (frameWidth / 2.0)
		
		minX = g2Object.MinX - padding
		maxX = g2Object.MaxX + padding
		minY = g2Object.MinY - padding
		maxY = g2Object.MaxY + padding
		#minZ = g2Object.MinZ - padding
		#maxZ = g2Object.MaxZ + padding
		minZ = g2Object.MinZ
		maxZ = g2Object.MaxZ
		
		diffX = maxX - minX
		diffY = maxY - minY
		diffZ = maxZ - minZ
		
		minBoxSide = (bracketLength * 2.0) + minBracketGap
		if diffX < minBoxSide:
			alteration = (minBoxSide - diffX) / 2.0
			minX -= alteration
			maxX += alteration
		if diffY < minBoxSide:
			alteration = (minBoxSide - diffY) / 2.0
			minY -= alteration
			maxY += alteration
		
		frameTopLeftBackX = getNewCube()
		frameTopLeftBackX.name = 'bb_%s_topleftbackx' % (baseName)
		frameTopLeftBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopLeftBackX.location = (minX + bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftBackX)
		
		frameTopLeftBackY = getNewCube()
		frameTopLeftBackY.name = 'bb_%s_topleftbacky' % (baseName)
		frameTopLeftBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopLeftBackY.location = (minX + bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftBackY)
		
		frameTopRightBackX = getNewCube()
		frameTopRightBackX.name = 'bb_%s_toprightbackx' % (baseName)
		frameTopRightBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopRightBackX.location = (maxX - bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightBackX)
		
		frameTopRightBackY = getNewCube()
		frameTopRightBackY.name = 'bb_%s_toprightbacky' % (baseName)
		frameTopRightBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopRightBackY.location = (maxX - bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightBackY)

		frameTopLeftFrontX = getNewCube()
		frameTopLeftFrontX.name = 'bb_%s_topleftfrontx' % (baseName)
		frameTopLeftFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopLeftFrontX.location = (minX + bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftFrontX)
		
		frameTopLeftFrontY = getNewCube()
		frameTopLeftFrontY.name = 'bb_%s_topleftfronty' % (baseName)
		frameTopLeftFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopLeftFrontY.location = (minX + bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftFrontY)
		
		frameTopRightFrontX = getNewCube()
		frameTopRightFrontX.name = 'bb_%s_toprightfrontx' % (baseName)
		frameTopRightFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopRightFrontX.location = (maxX - bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightFrontX)
		
		frameTopRightFrontY = getNewCube()
		frameTopRightFrontY.name = 'bb_%s_toprightfronty' % (baseName)
		frameTopRightFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopRightFrontY.location = (maxX - bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightFrontY)
		

		#WriteToFile(debugLogPath, 'Creating union of %s' % (elements))
		bpy.context.evaluated_depsgraph_get().update()
		
		bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
		bpy.ops.object.select_all(action='DESELECT')
		elements[0].select_set(True)
		bpy.context.view_layer.objects.active = elements[0]			
		
		for i in range(1, len(elements)):
			elements[i].select_set(True)
			bpy.ops.object.join()
		
		elements[0].name='bounding_box_%s_data' % (baseName)
		# mat = bpy.data.materials.new(name = 'm_bounding_box_%s' % (baseName))
		# mat.use_nodes= True
		# mat.blend_method = 'BLEND'
		# bsdfNode = None
		# for n in mat.node_tree.nodes:
			# if n.name == 'Principled BSDF':
				# bsdfNode = n
		# if bsdfNode:
			# bsdfNode.inputs['Base Color'].default_value = highlightColour
			# bsdfNode.inputs['Emission'].default_value = highlightColour
			# bsdfNode.inputs['Specular'].default_value = 0.0
			# bsdfNode.inputs['Alpha'].default_value = boxAlpha
		# if elements[0].data.materials:
			# elements[0].data.materials[0] = mat
		# else:
			# elements[0].data.materials.append(mat)
		
		bpy.ops.object.select_all(action='DESELECT')
		
		neFrame = g2Object.Clone()
		neFrame.BaseBlenderObject = elements[0]
		neFrame.ResetBoundingBox()
		neFrame.SetVertexBoundingBox(neFrame.BaseBlenderObject)
		neFrame.Name = elements[0].name
		newEntries.append(neFrame)
	
	return newEntries

def isAddedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("add8")
	individualAreas.append("aluka7")
	# individualAreas.append("aluka47")	# In the release version, this is the Fire Forge
	individualAreas.append("aluka48")
	individualAreas.append("bonus1")
	individualAreas.append("cathy50")
	individualAreas.append("cathy65")
	individualAreas.append("cathy67")	# deleted in 1999-06-01, then re-added in 1999-06-10
	individualAreas.append("cathy71")
	individualAreas.append("cathy72")
	individualAreas.append("cathy73")
	individualAreas.append("cathy74")
	individualAreas.append("cathy75")
	individualAreas.append("cathy76")
	individualAreas.append("chrono3")
	individualAreas.append("chrono4")
	individualAreas.append("chrono5")
	individualAreas.append("chrono6")
	individualAreas.append("chrono7")
	individualAreas.append("chrono8")
	individualAreas.append("chrono9")
	individualAreas.append("chrono10")
	individualAreas.append("chrono11")
	individualAreas.append("chrono12")
	individualAreas.append("chrono13")
	individualAreas.append("chrono14")
	individualAreas.append("chrono15")
	individualAreas.append("chrono16")
	individualAreas.append("chrono17")
	individualAreas.append("chrono18")
	individualAreas.append("chrono19")
	individualAreas.append("city19")
	individualAreas.append("city20")
	individualAreas.append("city21")
	individualAreas.append("clfpil4")
	individualAreas.append("fire8")
	individualAreas.append("huba18")
	individualAreas.append("huba19")
	individualAreas.append("huba20")
	individualAreas.append("mrlock14")
	individualAreas.append("nighta10")
	individualAreas.append("nighta11")
	individualAreas.append("nighta12")
	individualAreas.append("nighta13")
	individualAreas.append("nightb10")
	individualAreas.append("nightb11")
	individualAreas.append("oracle23")
	individualAreas.append("oracle24")
	individualAreas.append("out13")
	individualAreas.append("out14")
	# individualAreas.append("out15") # This is the Force Glyph room, renamed from aluka26
	individualAreas.append("out16")
	individualAreas.append("pillars10")
	individualAreas.append("skinnr18")
	individualAreas.append("soundg6")
	individualAreas.append("tower12")
	
	if areaNameLower in individualAreas:
		return True

	return False

def isMovedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("mrlock7")	# moved to the citadel
	individualAreas.append("mrlock8")	# moved to the citadel
	
	if areaNameLower in individualAreas:
		return True

	return False

def isSignificantlyChangedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("cathy19")	# there is a "cathy19" in the final version, but it's a shadow of what it was in the alpha builds
	individualAreas.append("pillars8")	# pillars8 in the alpha builds connects pillars9 to pillars7 (the alternate entrance to the Sanctuary)
	if areaNameLower in individualAreas:
		return True

	return False

def isDeletedArea(areaName):
	areaNameLower = areaName.lower()
	if areaNameLower.startswith("adda"):
		return True
	if areaNameLower.startswith("lair"):
		return True
	if areaNameLower.startswith("movie"):
		return True
	if areaNameLower.startswith("push"):
		return True
	if areaNameLower.startswith("retreat"):
		return True
	if areaNameLower.startswith("test"):
		return True
	if areaNameLower.startswith("undrct"):
		return True
	
	individualAreas = []
	individualAreas.append("aluka7")
	individualAreas.append("aluka13")
	#individualAreas.append("aluka16") # merged with adjacent area, not actually deleted
	individualAreas.append("aluka17")
	individualAreas.append("aluka18")
	#individualAreas.append("aluka26")	# This is the Force Glyph room, which became out15
	individualAreas.append("aluka35")	# this really is a deleted area, but it's just a short connecting passage
	individualAreas.append("aluka42")
	individualAreas.append("aluka47")	# deleted, then re-added
	individualAreas.append("cathy14")
	individualAreas.append("cathy16")
	individualAreas.append("cathy19")
	individualAreas.append("cathy20")
	individualAreas.append("cathy23")
	individualAreas.append("cathy26")
	individualAreas.append("cathy29")
	individualAreas.append("cathy34")
	individualAreas.append("cathy35")
	individualAreas.append("cathy41")
	individualAreas.append("cathy43")
	individualAreas.append("cathy44")
	individualAreas.append("cathy56")
	individualAreas.append("cathy63")
	individualAreas.append("cathy67")	# deleted in 1999-06-01, then re-added in 1999-06-10
	individualAreas.append("city10")
	individualAreas.append("city11")
	individualAreas.append("city12")
	individualAreas.append("city16")
	individualAreas.append("fill5")
	individualAreas.append("forge6")
	#individualAreas.append("huba12") # Fire Forge, moved to Drowned Abbey (where the Water Forge was), not actually deleted
	individualAreas.append("huba15") # merged with adjacent area, not actually deleted
	individualAreas.append("mrlock3")
	individualAreas.append("mrlock4")
	individualAreas.append("mrlock5")
	individualAreas.append("mrlock6")
	individualAreas.append("mrlock9")
	individualAreas.append("mrlock10")
	individualAreas.append("mrlock11")
	individualAreas.append("mrlock12")
	individualAreas.append("mrlock13")
	individualAreas.append("nighta9")
	individualAreas.append("nightb7")
	individualAreas.append("pillars7")
	individualAreas.append("soundg1")
	individualAreas.append("soundg2")
	#individualAreas.append("stone12")	# stone12 isn't present in later builds, but it was merged into one of the other Stone Glyph areas
	individualAreas.append("train10")

	if areaNameLower in individualAreas:
		return True

	return False

def addBracketBoxAroundObject(g2Object, objectNumber, frameWidth, bracketLength, fontSize, existingNewEntries, debugLogPath = "", minBracketGap = 0.5, textOffsetX = 20.0, textOffsetY = 10.0, textOffsetZ = 5.0, textPadding = 2.0, setPlacementOrder = True):
	newEntries = []
	global textOffsetNumNorthWest
	global textOffsetNumNorthEast
	global textOffsetNumSouthWest
	global textOffsetNumSouthEast
	frameColour = highlightColour
	for i in range(0, len(existingNewEntries)):
		newEntries.append(existingNewEntries[i])
	if g2Object.BaseBlenderObject:
		# Set the location information for all of the added objects to match the original object
		if setPlacementOrder:
			g2Object.ScriptPlacementOrder = objectNumber
			if 'scripted_placement_order' in g2Object.BaseBlenderObject.keys():
				g2Object.BaseBlenderObject['scripted_placement_order'] = g2Object.ScriptPlacementOrder
				WriteToFile(debugLogPath, "Updated scripted_placement_order property for object '%s' to %s" % (g2Object.Name, g2Object.ScriptPlacementOrder))
			else:
				WriteToFile(debugLogPath, "Couldn't find a scripted_placement_order property for object '%s'" % (g2Object.Name))
		baseName = g2Object.Name
		if baseName.startswith("area_"):
			baseName = baseName[5:]

		if isDeletedArea(baseName):
			frameColour = highlightColourDeletedArea
		
		if isSignificantlyChangedArea(baseName):
			frameColour = highlightColourDeletedArea

		awo = getAreaWithObjectsContainer(baseName)

		#bpy.context.evaluated_depsgraph_get().update()
		elements = []
		bracketPosOffsetLongAxis = (bracketLength / 2.0)
		bracketPosOffsetShortAxis = (frameWidth / 2.0)
		
		minX = g2Object.MinX
		maxX = g2Object.MaxX
		minY = g2Object.MinY
		maxY = g2Object.MaxY
		minZ = g2Object.MinZ
		maxZ = g2Object.MaxZ
		
		diffX = maxX - minX
		diffY = maxY - minY
		diffZ = maxZ - minZ
		
		minBoxSide = (bracketLength * 2.0) + minBracketGap
		if diffX < minBoxSide:
			alteration = (minBoxSide - diffX) / 2.0
			minX -= alteration
			maxX += alteration
		if diffY < minBoxSide:
			alteration = (minBoxSide - diffY) / 2.0
			minY -= alteration
			maxY += alteration
		if diffZ < minBoxSide:
			alteration = (minBoxSide - diffZ) / 2.0
			minZ -= alteration
			maxZ += alteration
		
		frameTopLeftBackX = getNewCube()
		frameTopLeftBackX.name = 'bb_%s_topleftbackx' % (baseName)
		frameTopLeftBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopLeftBackX.location = (minX + bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftBackX)
		
		frameTopLeftBackY = getNewCube()
		frameTopLeftBackY.name = 'bb_%s_topleftbacky' % (baseName)
		frameTopLeftBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopLeftBackY.location = (minX + bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftBackY)
		
		frameTopLeftBackZ = getNewCube()
		frameTopLeftBackZ.name = 'bb_%s_topleftbackz' % (baseName)
		frameTopLeftBackZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameTopLeftBackZ.location = (minX + bracketPosOffsetShortAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetLongAxis)
		elements.append(frameTopLeftBackZ)
		
		frameTopRightBackX = getNewCube()
		frameTopRightBackX.name = 'bb_%s_toprightbackx' % (baseName)
		frameTopRightBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopRightBackX.location = (maxX - bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightBackX)
		
		frameTopRightBackY = getNewCube()
		frameTopRightBackY.name = 'bb_%s_toprightbacky' % (baseName)
		frameTopRightBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopRightBackY.location = (maxX - bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightBackY)
		
		frameTopRightBackZ = getNewCube()
		frameTopRightBackZ.name = 'bb_%s_toprightbackz' % (baseName)
		frameTopRightBackZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameTopRightBackZ.location = (maxX - bracketPosOffsetShortAxis, maxY - bracketPosOffsetShortAxis, maxZ - bracketPosOffsetLongAxis)
		elements.append(frameTopRightBackZ)
		
		frameTopLeftFrontX = getNewCube()
		frameTopLeftFrontX.name = 'bb_%s_topleftfrontx' % (baseName)
		frameTopLeftFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopLeftFrontX.location = (minX + bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftFrontX)
		
		frameTopLeftFrontY = getNewCube()
		frameTopLeftFrontY.name = 'bb_%s_topleftfronty' % (baseName)
		frameTopLeftFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopLeftFrontY.location = (minX + bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopLeftFrontY)
		
		frameTopLeftFrontZ = getNewCube()
		frameTopLeftFrontZ.name = 'bb_%s_topleftfrontz' % (baseName)
		frameTopLeftFrontZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameTopLeftFrontZ.location = (minX + bracketPosOffsetShortAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetLongAxis)
		elements.append(frameTopLeftFrontZ)
		
		frameTopRightFrontX = getNewCube()
		frameTopRightFrontX.name = 'bb_%s_toprightfrontx' % (baseName)
		frameTopRightFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameTopRightFrontX.location = (maxX - bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightFrontX)
		
		frameTopRightFrontY = getNewCube()
		frameTopRightFrontY.name = 'bb_%s_toprightfronty' % (baseName)
		frameTopRightFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameTopRightFrontY.location = (maxX - bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, maxZ - bracketPosOffsetShortAxis)
		elements.append(frameTopRightFrontY)
		
		frameTopRightFrontZ = getNewCube()
		frameTopRightFrontZ.name = 'bb_%s_toprightfrontz' % (baseName)
		frameTopRightFrontZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameTopRightFrontZ.location = (maxX - bracketPosOffsetShortAxis, minY + bracketPosOffsetShortAxis, maxZ - bracketPosOffsetLongAxis)
		elements.append(frameTopRightFrontZ)
		
		frameBottomLeftBackX = getNewCube()
		frameBottomLeftBackX.name = 'bb_%s_bottomleftbackx' % (baseName)
		frameBottomLeftBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameBottomLeftBackX.location = (minX + bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomLeftBackX)
		
		frameBottomLeftBackY = getNewCube()
		frameBottomLeftBackY.name = 'bb_%s_bottomleftbacky' % (baseName)
		frameBottomLeftBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameBottomLeftBackY.location = (minX + bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomLeftBackY)
		
		frameBottomLeftBackZ = getNewCube()
		frameBottomLeftBackZ.name = 'bb_%s_bottomleftbackz' % (baseName)
		frameBottomLeftBackZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameBottomLeftBackZ.location = (minX + bracketPosOffsetShortAxis, maxY - bracketPosOffsetShortAxis, minZ + bracketPosOffsetLongAxis)
		elements.append(frameBottomLeftBackZ)
		
		frameBottomRightBackX = getNewCube()
		frameBottomRightBackX.name = 'bb_%s_bottomrightbackx' % (baseName)
		frameBottomRightBackX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameBottomRightBackX.location = (maxX - bracketPosOffsetLongAxis, maxY - bracketPosOffsetShortAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomRightBackX)
		
		frameBottomRightBackY = getNewCube()
		frameBottomRightBackY.name = 'bb_%s_bottomrightbacky' % (baseName)
		frameBottomRightBackY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameBottomRightBackY.location = (maxX - bracketPosOffsetShortAxis, maxY - bracketPosOffsetLongAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomRightBackY)
		
		frameBottomRightBackZ = getNewCube()
		frameBottomRightBackZ.name = 'bb_%s_bottomrightbackz' % (baseName)
		frameBottomRightBackZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameBottomRightBackZ.location = (maxX - bracketPosOffsetShortAxis, maxY - bracketPosOffsetShortAxis, minZ + bracketPosOffsetLongAxis)
		elements.append(frameBottomRightBackZ)
		
		frameBottomLeftFrontX = getNewCube()
		frameBottomLeftFrontX.name = 'bb_%s_bottomleftfrontx' % (baseName)
		frameBottomLeftFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameBottomLeftFrontX.location = (minX + bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomLeftFrontX)
		
		frameBottomLeftFrontY = getNewCube()
		frameBottomLeftFrontY.name = 'bb_%s_bottomleftfronty' % (baseName)
		frameBottomLeftFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameBottomLeftFrontY.location = (minX + bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomLeftFrontY)
		
		frameBottomLeftFrontZ = getNewCube()
		frameBottomLeftFrontZ.name = 'bb_%s_bottomleftfrontz' % (baseName)
		frameBottomLeftFrontZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameBottomLeftFrontZ.location = (minX + bracketPosOffsetShortAxis, minY + bracketPosOffsetShortAxis, minZ + bracketPosOffsetLongAxis)
		elements.append(frameBottomLeftFrontZ)
		
		frameBottomRightFrontX = getNewCube()
		frameBottomRightFrontX.name = 'bb_%s_bottomrightfrontx' % (baseName)
		frameBottomRightFrontX.dimensions = (bracketLength, frameWidth, frameWidth)
		frameBottomRightFrontX.location = (maxX - bracketPosOffsetLongAxis, minY + bracketPosOffsetShortAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomRightFrontX)
		
		frameBottomRightFrontY = getNewCube()
		frameBottomRightFrontY.name = 'bb_%s_bottomrightfronty' % (baseName)
		frameBottomRightFrontY.dimensions = (frameWidth, bracketLength, frameWidth)
		frameBottomRightFrontY.location = (maxX - bracketPosOffsetShortAxis, minY + bracketPosOffsetLongAxis, minZ + bracketPosOffsetShortAxis)
		elements.append(frameBottomRightFrontY)
		
		frameBottomRightFrontZ = getNewCube()
		frameBottomRightFrontZ.name = 'bb_%s_bottomrightfrontz' % (baseName)
		frameBottomRightFrontZ.dimensions = (frameWidth, frameWidth, bracketLength)
		frameBottomRightFrontZ.location = (maxX - bracketPosOffsetShortAxis, minY + bracketPosOffsetShortAxis, minZ + bracketPosOffsetLongAxis)
		elements.append(frameBottomRightFrontZ)
		
		
		
		textPositionX = g2Object.GetMidX()
		textPositionY = g2Object.GetMidY()
		textPositionZ = maxZ + textOffsetZ
		
		#if textPositionX <= 0.0:
		#	textPositionX = g2Object.MinX - (textOffsetX + textPadding)
		#else:
		#	textPositionX = g2Object.MaxX + (textOffsetX + textPadding)
		#if textPositionY <= 0.0:
		#	textPositionY = g2Object.MinY - (textOffsetY + textPadding)
		#else:
		#	textPositionY = g2Object.MaxY + (textOffsetY + textPadding)
		
		# textOffsetMultipliers
		textOffsetXReal = textOffsetX
		textOffsetYReal = textOffsetY
		if textPositionX <= 0.0:
			# Southwest quadrant
			if textPositionY <= 0.0:
				textOffsetXReal = textOffsets[textOffsetNumSouthWest][0]
				textOffsetYReal = textOffsets[textOffsetNumSouthWest][1]
				textOffsetNumSouthWest = (textOffsetNumSouthWest + 1) % len(textOffsets)
			# Northwest quadrant
			else:
				textOffsetXReal = textOffsets[textOffsetNumNorthWest][0]
				textOffsetYReal = textOffsets[textOffsetNumNorthWest][1]
				textOffsetNumNorthWest = (textOffsetNumNorthWest + 1) % len(textOffsets)
		else:
			# Southeast quadrant
			if textPositionY <= 0.0:
				textOffsetXReal = textOffsets[textOffsetNumSouthEast][0]
				textOffsetYReal = textOffsets[textOffsetNumSouthEast][1]
				textOffsetNumSouthEast = (textOffsetNumSouthEast + 1) % len(textOffsets)
			# Northeast quadrant
			else:
				textOffsetXReal = textOffsets[textOffsetNumNorthEast][0]
				textOffsetYReal = textOffsets[textOffsetNumNorthEast][1]
				textOffsetNumNorthEast = (textOffsetNumNorthEast + 1) % len(textOffsets)
		
		if textPositionX <= 0.0:
			textPositionX = minX - (textOffsetXReal + textPadding)
		else:
			textPositionX = maxX + (textOffsetXReal + textPadding)
		if textPositionY <= 0.0:
			textPositionY = minY - (textOffsetYReal + textPadding)
		else:
			textPositionY = maxY + (textOffsetYReal + textPadding)
		
		textEntries = drawTextCenteredAtLocation(g2Object, baseName, textFont, fontSize, baseName, (textPositionX, textPositionY, textPositionZ), frameColour)
		for i in range(0, len(textEntries)):
			#textEntries[i].SetVertexBoundingBoxFromOtherObject(g2Object)
			textEntries[i].ScriptPlacementOrder = g2Object.ScriptPlacementOrder
			textEntries[i].DistanceFromOrigin = g2Object.DistanceFromOrigin
			newEntries.append(textEntries[i])
			awo.objects.link(textEntries[i].BaseBlenderObject)
		
		textBracketEntries = addBracketFrameAroundObject(textEntries[0], "%s_textframe" % (baseName), frameWidth, bracketLength, padding = textPadding, debugLogPath = debugLogPath)
		for i in range(0, len(textBracketEntries)):
			elements.append(textBracketEntries[i].BaseBlenderObject)
			#newEntries.append(textBracketEntries[i])
			#awo.objects.link(textBracketEntries[i].BaseBlenderObject)

		#calculate the length of the connecting bar
		cbEndX = textBracketEntries[0].MaxX
		cbEndY = textBracketEntries[0].MaxY
		cbEndZ = textBracketEntries[0].MinZ
		cbStartX = minX
		cbStartY = minY
		cbStartZ = maxZ
		#cbDiffX = cbEndX - cbStartX
		#cbDiffY = cbEndY - cbStartY
		cbDiffX = cbStartX - cbEndX
		cbDiffY = cbStartY - cbEndY
		cbDiffZ = cbEndZ - cbStartZ
		if textPositionX > 0.0:
			cbEndX = textBracketEntries[0].MinX
			cbStartX = maxX
			cbDiffX = cbEndX - cbStartX
			#cbDiffX = cbStartX - cbEndX
		if textPositionY > 0.0:
			cbEndY = textBracketEntries[0].MinY 
			cbStartY = maxY
			cbDiffY = cbEndY - cbStartY
			#cbDiffY = cbStartY - cbEndY

		cbLength = math.pow(math.pow((cbDiffX), 2.0) + math.pow((cbDiffY), 2.0) + math.pow((cbDiffZ), 2.0), 0.5) + (frameWidth * 3.0)
		
		#cbYaw = math.atan2(cbDiffZ, cbDiffX)
		#cbPitch = (math.atan2(math.sqrt(math.pow(cbDiffZ, 2.0) + math.pow(cbDiffX, 2.0)), cbDiffY) + math.pi)
		#cbYaw = math.atan2(cbDiffZ, cbDiffX) - (math.pi / 4.0)
		#cbPitch = (math.atan2(cbDiffY, math.sqrt(math.pow(cbDiffZ, 2.0) + math.pow(cbDiffX, 2.0))))
		#cbYaw = math.atan2(cbDiffX, cbDiffZ)
		#cbPitch = (math.atan2(cbDiffY, math.sqrt(math.pow(cbDiffZ, 2.0) + math.pow(cbDiffX, 2.0))))
		#sqrtXY = math.sqrt(math.pow(cbDiffX, 2.0) + math.pow(cbDiffY, 2.0))
		#l = math.atan(cbDiffY / cbDiffX)
		#phi = math.atan(cbDiffZ / sqrtXY)
		#if cbDiffX < 0:
		#	phi = phi + math.pi
		#if cbDiffZ < 0:
		#	l *= -1.0
		
		# this is perfect for Y and X both positive
		cbYaw = math.atan2(cbDiffX, cbDiffY) * -1.0
		cbPitch = math.atan2(cbDiffZ, math.sqrt(math.pow(cbDiffX, 2.0) + math.pow(cbDiffY, 2.0)))
		cbRoll = 0.0
		
		# flipRotation = False
		# if textPositionX <= 0.0 and textPositionY > 0.0:
			# flipRotation = True
		# if textPositionX > 0.0 and textPositionY <= 0.0:
			# flipRotation = True
		# if flipRotation:
			# cbPitch *= -1.0
			# cbYaw *= -1.0

		if textPositionY > 0:
			if textPositionX <= 0.0:
				cbYaw *= -1.0
		else:
			cbPitch *= -1.0
			if textPositionX > 0.0:
				cbYaw *= -1.0
		#if textPositionY <= 0.0:
		#	cbPitch *= -1.0
		
		# if textPositionX > 0.0:
			# #cbPitch *= -1.0
			# cbPitch = (math.pi * 2.0) - cbPitch
			# if textPositionY <= 0.0:
				# #cbYaw *= -1.0
				# cbYaw = (math.pi * 2.0) - cbYaw
		# else:
			# if textPositionY > 0.0:
				# #cbYaw *= -1.0
				# cbYaw = (math.pi * 2.0) - cbYaw
		
		#if textPositionX <= 0.0:
		#	if textPositionY > 0.0:
		#		cbPitch = (math.pi * 2.0) - cbPitch
		#		cbYaw = (math.pi * 2.0) - cbYaw
					
		connectingBar = getNewCube()
		connectingBar.name = 'bb_%s_connectingbar' % (baseName)
		#connectingBar.dimensions = (frameWidth, cbLength, frameWidth)
		connectingBar.dimensions = (frameWidth, cbLength, frameWidth)
		# close, but looks like rotation has the two angles used for each other's purpose
		#connectingBar.rotation_euler[1] = cbPitch
		#connectingBar.rotation_euler[2] = cbYaw
		# mirrored backwards along the Y axis:
		#connectingBar.rotation_euler[2] = cbPitch
		#connectingBar.rotation_euler[1] = cbYaw
		# correct yaw, but incorrect pitch:
		#connectingBar.rotation_euler[0] = cbPitch
		#connectingBar.rotation_euler[2] = cbYaw
		try:
			bpy.context.scene.collection.objects.link(connectingBar)
		except Exception as e:
			WriteToFile(debugLogPath, "Error linking connecting bar '%s': %s" % (connectingBar.name, e))
		bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
		bpy.ops.object.select_all(action='DESELECT')
		connectingBar.select_set(state = True)
		bpy.context.view_layer.objects.active = connectingBar
		connectingBar.rotation_euler[2] = cbYaw		
		#bpy.ops.object.transform_apply(location = False, rotation = True, scale = False)
		connectingBar.rotation_euler[0] = cbPitch
		#bpy.ops.object.transform_apply(location = False, rotation = True, scale = False)		
		connectingBar.rotation_euler[1] = cbRoll
		bpy.ops.object.transform_apply(location = False, rotation = True, scale = False)		
		bpy.ops.object.select_all(action='DESELECT')

		if textPositionX <= 0.0:
			connectingBar.location[0] = cbStartX - (cbDiffX / 2.0)
		else:
			connectingBar.location[0] = cbStartX + (cbDiffX / 2.0)			
		if textPositionY <= 0.0:
			connectingBar.location[1] = cbStartY - (cbDiffY / 2.0)
		else:
			connectingBar.location[1] = cbStartY + (cbDiffY / 2.0)
		connectingBar.location[2] = cbStartZ + (cbDiffZ / 2.0)
		elements.append(connectingBar)

		#WriteToFile(debugLogPath, 'Creating union of %s' % (elements))
		bpy.context.evaluated_depsgraph_get().update()
		
		bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
		bpy.ops.object.select_all(action='DESELECT')
		elements[0].select_set(True)
		bpy.context.view_layer.objects.active = elements[0]			
		
		for i in range(1, len(elements)):
			elements[i].select_set(True)
			bpy.ops.object.join()
		
		elements[0].name='bounding_box_%s_data' % (baseName)
		mat = bpy.data.materials.new(name = 'm_bounding_box_%s' % (baseName))
		mat.use_nodes= True
		mat.blend_method = 'BLEND'
		bsdfNode = None
		for n in mat.node_tree.nodes:
			if n.name == 'Principled BSDF':
				bsdfNode = n
		if bsdfNode:
			bsdfNode.inputs['Base Color'].default_value = frameColour
			bsdfNode.inputs['Emission'].default_value = frameColour
			bsdfNode.inputs['Specular'].default_value = 0.0
			bsdfNode.inputs['Alpha'].default_value = boxAlpha
		if elements[0].data.materials:
			elements[0].data.materials[0] = mat
		else:
			elements[0].data.materials.append(mat)
		
		bpy.ops.object.select_all(action='DESELECT')
		
		awo.objects.link(elements[0])
		
		neFrame = g2Object.Clone()
		neFrame.BaseBlenderObject = elements[0]
		neFrame.Name = elements[0].name
		#neFrame.SetVertexBoundingBoxFromOtherObject(g2Object)
		neFrame.ScriptPlacementOrder = g2Object.ScriptPlacementOrder
		neFrame.DistanceFromOrigin = g2Object.DistanceFromOrigin
		newEntries.append(neFrame)
		

	#for i in range(0, len(newEntries)):
	#	newEntries[i].SetVertexBoundingBoxFromOtherObject(g2Object)
	#	newEntries[i].ScriptPlacementOrder = g2Object.ScriptPlacementOrder
	#	newEntries[i].DistanceFromOrigin = g2Object.DistanceFromOrigin
	
	return newEntries

def addBracketBoxToObjectsInList(objectList, debugLogPath = "", setPlacementOrder = True):
	frameWidth = 0.5
	bracketLength = 3.0
	fontSize = 6.0
	newEntries = []
	WriteToFile(debugLogPath, "Adding bracket boxes to areas")
	eNum = 0
	for g2o in objectList:
		if "_obj_" not in g2o.Name:
			newEntries = addBracketBoxAroundObject(g2o, eNum, frameWidth, bracketLength, fontSize, newEntries, debugLogPath = debugLogPath, setPlacementOrder = setPlacementOrder)
		eNum += 1
		if eNum == len(objectList) or (eNum % 10) == 0:
			msg = "%s\t%s/%s objects processed" % (time.time(), eNum, len(objectList))
			WriteToFile(debugLogPath, msg)

	for ne in newEntries:
		objectList.append(ne)

	return objectList
	
def setObjectPlacementOrderToParent(objectList, debugLogPath = ""):
	eNum = 0
	WriteToFile(debugLogPath, "Setting object placement orders to match their parent areas")
	for g2o in objectList:
		if "_obj_" in g2o.Name:
			areaName = g2o.Name.split('_')[0]
			areaBlenderObjectName = "area_%s" % (areaName)
			if areaBlenderObjectName in bpy.data.objects.keys():
				abo = bpy.data.objects[areaBlenderObjectName]
				if 'scripted_placement_order' in abo.keys():
					g2o.BaseBlenderObject['scripted_placement_order'] = abo['scripted_placement_order']
					g2o.ScriptPlacementOrder = abo['scripted_placement_order']
				else:
					WriteToFile(debugLogPath, "Parent area '%s' is missing the scripted_placement_order property" % (areaBlenderObjectName))
				if 'distance_from_origin' in abo.keys():
					g2o.BaseBlenderObject['distance_from_origin'] = abo['distance_from_origin']
					g2o.DistanceFromOrigin = abo['distance_from_origin']
				else:
					WriteToFile(debugLogPath, "Parent area '%s' is missing the distance_from_origin property" % (areaBlenderObjectName))
			else:
				WriteToFile(debugLogPath, "Could not find parent area '%s' to set object properties" % (areaBlenderObjectName))
		eNum += 1
		if eNum == len(objectList) or (eNum % 10) == 0:
			msg = "%s\t%s/%s objects processed" % (time.time(), eNum, len(objectList))
			WriteToFile(debugLogPath, msg)

	return objectList



# https://docs.python.org/3/howto/sorting.html
# multisort(list(student_objects), (('grade', True), ('age', False)))
def multisort(xs, specs):
	if len(specs) > 1 and len(specs[0]) > 1:
		for key, reverse in reversed(specs):
			xs.sort(key=attrgetter(key), reverse=reverse)
	else:
		xs.sort(key=attrgetter(specs[0]), reverse=specs[1])
	return xs


def setMaterialAlphaKeyframe(obj, frameNumber, alpha):
	bpy.context.scene.frame_set(frameNumber)
	for mat in obj.data.materials:
		for n in mat.node_tree.nodes:
			for nInput in n.inputs:
				if 'Alpha' in nInput.name:
					nInput.default_value = alpha
					nInput.keyframe_insert(data_path="default_value", index=-1)	

dlPath = 'X:\\Legacy_of_Kain\\Soul_Reaver_Alpha\\ObjectArrangeDebugLog-Spiral_Outward.txt'

# place objects one by one? (vs with their parent area)
placeObjectsSimultaneously = True

# How far to move, and in what direction, areas and objects
travelDistanceAreas = 1200.0
travelDistanceObjects = 1200.0
animationPositionOffsetAreasBase = (0.0, 0.0, travelDistanceAreas)
animationPositionOffsetObjectsBase = (0.0, 0.0, travelDistanceObjects)
# Number of frames to move elements over between their start and end positions
animationFrameCountAreas = 30
animationFrameCountObjects = 30
animationFrameCountTextFadeout = 30
animationFrameCountHoldAtEnd = 30
# Number of frames to offset moving each element by
# For example, if animationFrameCountAreas and animationFrameCountObjects are both 30, 
# animationPerElementFrameOffsetAreas is 15, and animationPerElementFrameOffsetObjects is 1, 
# there are three areas (A, B, and C), and 30 objects, then:
# Area A will begin moving at frame 1 and end at frame 30
# Area B will begin moving at frame 15 and end at frame 45
# Area C will begin moving at frame 31 and end at frame 60
# One object will begin moving at each frame starting at 61 and end 30 frames after it 
# starts
# For individual levels
#animationPerElementFrameOffsetAreas = 15
#animationPerElementFrameOffsetObjects = 5
# number of frames to pause when switching from area => object placement
#pauseFramesAreaToObject = 5
pauseFramesAreaToObject = 0
# number of frames to pause when switching from object => area placement
#pauseFramesObjectToArea = 30
#pauseFramesObjectToArea = 10
pauseFramesObjectToArea = 0

# for the world map
animationPerElementFrameOffsetAreas = 5
animationPerElementFrameOffsetObjects = 1
# number of frames to pause when switching from area => object placement
#pauseFramesAreaToObject = 0
# number of frames to pause when switching from object => area placement
#pauseFramesObjectToArea = 0

currentFrameNumber = 1

g2Elements = getSortableObjectList(debugLogPath = dlPath)

# only place objects that match a regex, and leave all others in place:
# g2Elements = getSortableObjectList(debugLogPath = dlPath, areaNameWhitelistRegex = 'undrct[0-9]+')
#g2Elements = addBoxesToObjectsInList(g2Elements, debugLogPath = dlPath)
# Lowest to highest, farthest to nearest, left to right (For tall areas)
#g2ElementsSorted = multisort(g2Elements, (('MaxZ', False), ('MinZ', False), ('MinY', True), ('MaxY', True), ('MinX', False), ('MaxX', False), ('ScriptPlacementOrder', False)))
# Lowest to highest, farthest to nearest, left to right (alt, not very good)
#g2ElementsSorted = multisort(g2Elements, (('MaxZ', False), ('MinY', True), ('MinX', False), ('MaxX', False), ('MinZ', False), ('MaxY', True)))
# Farthest to nearest, lowest to highest, left to right (shallow but broad areas)
g2ElementsSorted = multisort(g2Elements, (('MinY', True), ('MaxY', True), ('MaxZ', False), ('MinZ', False), ('MinX', False), ('MaxX', False), ('ScriptPlacementOrder', False)))
# expand outward from the centermost object
#g2ElementsSorted = multisort(g2Elements, (('AbsoluteMaxZ', False), ('AbsoluteMaxY', False), ('AbsoluteMaxX', False), ('AbsoluteMinZ', False), ('AbsoluteMinY', False), ('AbsoluteMinX', False), ('ScriptPlacementOrder', False)))

# Add bracket boxes if so desired
g2ElementsSorted = addBracketBoxToObjectsInList(g2ElementsSorted, debugLogPath = dlPath, setPlacementOrder = True)

# Use the cyclic placement mode for spiral-outward arrangement
#useCyclicStartOffset = True
useCyclicStartOffset = False

#WriteToFile(dlPath, "%s" % (g2ElementsSorted))
g2ElementsSorted = setObjectPlacementOrderToParent(g2ElementsSorted, debugLogPath = dlPath)
g2ElementsSorted.sort(key=attrgetter('ScriptPlacementOrder'), reverse=False)

if len(g2Elements) == 0:
	WriteToFile(dlPath, "Error: did not find any G2 elements in the list of Blender objects in the current project")
else:
	# Raster-style arrangement
	# 	Sort in the following order:
	#		Lowest Z position to highest
	#		Most distant Y position to closest
	#		Left to right on the X axis
	#g2ElementsSorted = multisort(g2Elements, (('MinZ', False), ('MaxZ', False), ('MinY', True), ('MaxY', True), ('MinX', True), ('MaxX', True), ('ScriptPlacementOrder', False)))
	#g2ElementsSorted = multisort(g2Elements, (('MaxZ', False), ('MinZ', False), ('MinY', True), ('MaxY', True), ('MinX', False), ('MaxX', False), ('ScriptPlacementOrder', False)))
	# In-order-added arrangement
	#g2ElementsSorted = g2Elements
	#g2ElementsSorted.sort(key=attrgetter('ScriptPlacementOrder'), reverse=False)
	#g2ElementsSorted = multisort(g2Elements, (('ScriptPlacementOrder', False)))
	# simultaneous placement of all areas at each given distance from the point of origin
	#g2ElementsSorted = multisort(g2Elements, (('DistanceFromOrigin', False)))
	#g2ElementsSorted = g2Elements
	#g2ElementsSorted.sort(key=attrgetter('DistanceFromOrigin'), reverse=False)
	#placementMode = "all_at_same_value:DistanceFromOrigin"
	placementMode = "all_at_same_value:ScriptPlacementOrder"
	
	lastElementType = "none"
	
	if placementMode == "one_by_one":
		for g2o in g2ElementsSorted:
			g2obj = None
			try:
				g2obj = bpy.data.objects[g2o.Name]
			except Exception as e:
				g2obj = None
				WriteToFile(dlPath, "Error: could not find the object '%s' in the Blender project")
			if g2obj:
				WriteToFile(dlPath, 'Placing object %s with properties: %s' % (g2o.Name, g2o.GetPropertiesString()))
				animationPositionOffsetAreas = animationPositionOffsetAreasBase
				animationPositionOffsetObjects = animationPositionOffsetObjectsBase
				gotElementType = False
				if g2o.G2Type == "area" or g2o.G2Type == "object":
					gotElementType = True
				if g2o.G2Type == "area":
					if lastElementType == "area":
						currentFrameNumber += animationPerElementFrameOffsetAreas
					if lastElementType == "object":
						currentFrameNumber += pauseFramesObjectToArea
				if g2o.G2Type == "object" or not gotElementType:
					if lastElementType == "area":
						currentFrameNumber += pauseFramesAreaToObject
					if lastElementType == "object":
						currentFrameNumber += animationPerElementFrameOffsetObjects
				if g2o.G2Type == "area":
					bpy.context.scene.frame_set(currentFrameNumber + animationFrameCountAreas)
				if g2o.G2Type == "object" or not gotElementType:
					bpy.context.scene.frame_set(currentFrameNumber + animationFrameCountObjects)
				g2obj.keyframe_insert(data_path="location", index=-1)
				#bpy.context.scene.frame_set(currentFrameNumber)
				bpy.context.scene.frame_set(0)
				if g2o.G2Type == "area":
					g2obj.location[0] += animationPositionOffsetAreas[0]
					g2obj.location[1] += animationPositionOffsetAreas[1]
					g2obj.location[2] += animationPositionOffsetAreas[2]
				if g2o.G2Type == "object" or not gotElementType:
					g2obj.location[0] += animationPositionOffsetObjects[0]
					g2obj.location[1] += animationPositionOffsetObjects[1]
					g2obj.location[2] += animationPositionOffsetObjects[2]
				g2obj.keyframe_insert(data_path="location", index=-1)
				bpy.context.scene.frame_set(currentFrameNumber)
				if g2o.G2Type == "area":
					g2obj.location[0] += animationPositionOffsetAreas[0]
					g2obj.location[1] += animationPositionOffsetAreas[1]
					g2obj.location[2] += animationPositionOffsetAreas[2]
				if g2o.G2Type == "object" or not gotElementType:
					g2obj.location[0] += animationPositionOffsetObjects[0]
					g2obj.location[1] += animationPositionOffsetObjects[1]
					g2obj.location[2] += animationPositionOffsetObjects[2]
				g2obj.keyframe_insert(data_path="location", index=-1)
				
				lastElementType = g2o.G2Type
	if "all_at_same_value" in placementMode:
		baseCounter = 0
		lastCounterValue = 0
		currentCounterValue = 0
		objectsInARowCount = 0
		debugCounter = 0
		
		cyclicPlacementCounterNWPlusZ = 0
		cyclicPlacementCounterNWNegZ = 0
		cyclicPlacementCounterNEPlusZ = 0
		cyclicPlacementCounterNENegZ = 0
		cyclicPlacementCounterSWPlusZ = 0
		cyclicPlacementCounterSWNegZ = 0
		cyclicPlacementCounterSEPlusZ = 0
		cyclicPlacementCounterSENegZ = 0
		
		cyclicPlacementValuesNWPlusZ = [ (travelDistanceAreas * -1.0, 0.0, 0.0), (0.0, travelDistanceAreas, 0.0), (0.0, 0.0, travelDistanceAreas) ]
		cyclicPlacementValuesNWNegZ = [ (travelDistanceAreas * -1.0, 0.0, 0.0), (0.0, travelDistanceAreas, 0.0), (0.0, 0.0, travelDistanceAreas * -1.0) ]
		cyclicPlacementValuesNEPlusZ = [ (travelDistanceAreas, 0.0, 0.0), (0.0, travelDistanceAreas, 0.0), (0.0, 0.0, travelDistanceAreas) ]
		cyclicPlacementValuesNENegZ = [ (travelDistanceAreas, 0.0, 0.0), (0.0, travelDistanceAreas, 0.0), (0.0, 0.0, travelDistanceAreas * -1.0) ]
		cyclicPlacementValuesSWPlusZ = [ (travelDistanceAreas * -1.0, 0.0, 0.0), (0.0, travelDistanceAreas * -1.0, 0.0), (0.0, 0.0, travelDistanceAreas) ]
		cyclicPlacementValuesSWNegZ = [ (travelDistanceAreas * -1.0, 0.0, 0.0), (0.0, travelDistanceAreas * -1.0, 0.0), (0.0, 0.0, travelDistanceAreas * -1.0) ]
		cyclicPlacementValuesSEPlusZ = [ (travelDistanceAreas, 0.0, 0.0), (0.0, travelDistanceAreas * -1.0, 0.0), (0.0, 0.0, travelDistanceAreas) ]
		cyclicPlacementValuesSENegZ = [ (travelDistanceAreas, 0.0, 0.0), (0.0, travelDistanceAreas * -1.0, 0.0), (0.0, 0.0, travelDistanceAreas * -1.0) ]
		
		for g2o in g2ElementsSorted:
			g2obj = None
			try:
				g2obj = bpy.data.objects[g2o.Name]
			except Exception as e:
				g2obj = None
				WriteToFile(dlPath, "Error: could not find the object '%s' in the Blender project")
			if g2obj:
				WriteToFile(dlPath, 'Placing object %s with properties: %s' % (g2o.Name, g2o.GetPropertiesString()))
				distanceFromOrigin = g2o.DistanceFromOrigin
				scriptPlacementOrder = g2o.ScriptPlacementOrder
				animationPositionOffsetAreas = animationPositionOffsetAreasBase
				animationPositionOffsetObjects = animationPositionOffsetObjectsBase
				
				if useCyclicStartOffset:
					offset = animationPositionOffsetAreasBase
					if g2o.MaxY > 0:
						if g2o.MaxX > 0:
							if g2o.MaxZ > 0:
								offset = cyclicPlacementValuesNEPlusZ[cyclicPlacementCounterNEPlusZ]
								cyclicPlacementCounterNEPlusZ = (cyclicPlacementCounterNEPlusZ + 1) % len(cyclicPlacementValuesNEPlusZ)
							else:
								offset = cyclicPlacementValuesNENegZ[cyclicPlacementValuesNENegZ]
								cyclicPlacementCounterNENegZ = (cyclicPlacementCounterNENegZ + 1) % len(cyclicPlacementValuesNENegZ)
						else:
							if g2o.MaxZ > 0:
								offset = cyclicPlacementValuesNWPlusZ[cyclicPlacementCounterNWPlusZ]
								cyclicPlacementCounterNWPlusZ = (cyclicPlacementCounterNWPlusZ + 1) % len(cyclicPlacementValuesNWPlusZ)
							else:
								offset = cyclicPlacementValuesNWNegZ[cyclicPlacementCounterNWNegZ]
								cyclicPlacementCounterNWNegZ = (cyclicPlacementCounterNWNegZ + 1) % len(cyclicPlacementValuesNWNegZ)
					else:
						if g2o.MaxX > 0:
							if g2o.MaxZ > 0:
								offset = cyclicPlacementValuesSEPlusZ[cyclicPlacementCounterSEPlusZ]
								cyclicPlacementCounterSEPlusZ = (cyclicPlacementCounterSEPlusZ + 1) % len(cyclicPlacementValuesSEPlusZ)
							else:
								offset = cyclicPlacementValuesSENegZ[cyclicPlacementCounterSENegZ]
								cyclicPlacementCounterSENegZ = (cyclicPlacementCounterSENegZ + 1) % len(cyclicPlacementValuesSENegZ)
						else:
							if g2o.MaxZ > 0:
								offset = cyclicPlacementValuesSWPlusZ[cyclicPlacementCounterSWPlusZ]
								cyclicPlacementCounterSWPlusZ = (cyclicPlacementCounterSWPlusZ + 1) % len(cyclicPlacementValuesSWPlusZ)
							else:
								offset = cyclicPlacementValuesSWNegZ[cyclicPlacementCounterSWNegZ]
								cyclicPlacementCounterSWNegZ = (cyclicPlacementCounterSWNegZ + 1) % len(cyclicPlacementValuesSWNegZ)
					animationPositionOffsetAreas = offset
					animationPositionOffsetObjects = offset
								
				# if the object has a parent, *and* the option to place objects simultaneously is set, try to use the parent values
				# if placeObjectsSimultaneously:
					# if g2obj.parent and g2obj.parent != "":
						# if 'scripted_placement_order' in g2obj.keys():
							# scriptPlacementOrder = int(g2obj['scripted_placement_order'])
							# WriteToFile(dlPath, "object: %s, using parent object's script placement order of %s instead of own value (%s)" % (g2o.Name, scriptPlacementOrder, g2o.ScriptPlacementOrder))
						# if 'distance_from_origin' in g2obj.keys():
							# distanceFromOrigin = int(g2obj['distance_from_origin'])
							# WriteToFile(dlPath, "object: %s, using parent object's distance from origin of %s instead of own value (%s)" % (g2o.Name, distanceFromOrigin, g2o.DistanceFromOrigin))
				if "DistanceFromOrigin" in placementMode:
					currentCounterValue = distanceFromOrigin
				if "ScriptPlacementOrder" in placementMode:
					currentCounterValue = scriptPlacementOrder
				#if placeObjectsSimultaneously:
				#	baseCounter = currentCounterValue
				#else:
				WriteToFile(dlPath, "object: %s, baseCounter: %s, currentCounterValue: %s, lastCounterValue: %s" % (g2o.Name, baseCounter, currentCounterValue, lastCounterValue))
				if currentCounterValue != lastCounterValue:
					baseCounter += 1
					WriteToFile(dlPath, "baseCounter updated to %s" % (baseCounter))
				gotElementType = False
				currentFrameNumber = baseCounter * animationPerElementFrameOffsetAreas
				if g2o.G2Type == "area" or g2o.G2Type == "object":
					gotElementType = True
				if g2o.G2Type == "object" or not gotElementType:
					if lastElementType == "object":
						objectsInARowCount += 1
					else:
						objectsInARowCount = 0
				else:
					objectsInARowCount = 0
				if placeObjectsSimultaneously:
					objectsInARowCount = 0
				if g2o.G2Type == "area":
					bpy.context.scene.frame_set(currentFrameNumber + animationFrameCountAreas)
				if g2o.G2Type == "object" or not gotElementType:
					bpy.context.scene.frame_set(currentFrameNumber + animationFrameCountObjects + (objectsInARowCount * animationPerElementFrameOffsetObjects))
				# set keyframes for final state of object
				WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
				g2obj.keyframe_insert(data_path="location", index=-1)
				startAlpha = boxAlpha
				fadeOut = False
				if "bounding_box" in g2o.Name:
					fadeOut = True
					if "font_object1" in g2o.Name:
						startAlpha = outlineAlpha
					if "font_object2" in g2o.Name:
						startAlpha = textAlpha
					currentFrame = bpy.context.scene.frame_current
					WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
					setMaterialAlphaKeyframe(g2o.BaseBlenderObject, currentFrame + animationFrameCountTextFadeout, 0.0)
					bpy.context.scene.frame_set(currentFrame)
				
				# set keyframes for frame 0 state of object
				bpy.context.scene.frame_set(0)
				if g2o.G2Type == "area":
					g2obj.location[0] += animationPositionOffsetAreas[0]
					g2obj.location[1] += animationPositionOffsetAreas[1]
					g2obj.location[2] += animationPositionOffsetAreas[2]
				if g2o.G2Type == "object" or not gotElementType:
					#bpy.context.scene.frame_set(currentFrameNumber + (objectsInARowCount * animationPerElementFrameOffsetObjects))
					g2obj.location[0] += animationPositionOffsetObjects[0]
					g2obj.location[1] += animationPositionOffsetObjects[1]
					g2obj.location[2] += animationPositionOffsetObjects[2]
				WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
				g2obj.keyframe_insert(data_path="location", index=-1)
				if fadeOut:
					currentFrame = bpy.context.scene.frame_current
					WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
					setMaterialAlphaKeyframe(g2o.BaseBlenderObject, currentFrame, startAlpha)
					bpy.context.scene.frame_set(currentFrame)
					
				# set keyframes for frame where object starts moving
				bpy.context.scene.frame_set(currentFrameNumber)
				if g2o.G2Type == "area":
					g2obj.location[0] += animationPositionOffsetAreas[0]
					g2obj.location[1] += animationPositionOffsetAreas[1]
					g2obj.location[2] += animationPositionOffsetAreas[2]
				if g2o.G2Type == "object" or not gotElementType:
					bpy.context.scene.frame_set(currentFrameNumber + (objectsInARowCount * animationPerElementFrameOffsetObjects))
					g2obj.location[0] += animationPositionOffsetObjects[0]
					g2obj.location[1] += animationPositionOffsetObjects[1]
					g2obj.location[2] += animationPositionOffsetObjects[2]
				WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
				g2obj.keyframe_insert(data_path="location", index=-1)
				if fadeOut:
					currentFrame = bpy.context.scene.frame_current
					WriteToFile(dlPath, "Setting keyframe for object '%s' at frame %s" % (g2obj.name, bpy.context.scene.frame_current))
					setMaterialAlphaKeyframe(g2o.BaseBlenderObject, currentFrame, startAlpha)					#setMaterialAlphaKeyframe(g2o.BaseBlenderObject, currentFrame + animationFrameCountTextFadeout, 0.0)
					bpy.context.scene.frame_set(currentFrame)
				lastElementType = g2o.G2Type
				lastCounterValueUpdated = False
				if "DistanceFromOrigin" in placementMode:
					lastCounterValueUpdated = True
					lastCounterValue = distanceFromOrigin			
				if "ScriptPlacementOrder" in placementMode:
					lastCounterValueUpdated = True
					lastCounterValue = scriptPlacementOrder
				if lastCounterValueUpdated:
					WriteToFile(dlPath, "lastCounterValue updated to %s" % (lastCounterValue))
			debugCounter += 1
			#if debugCounter > 20:
			#	break

maxFrames = pauseFramesObjectToArea
if pauseFramesAreaToObject > maxFrames:
	maxFrames = pauseFramesAreaToObject
if animationFrameCountAreas > maxFrames:
	maxFrames = animationFrameCountAreas
if animationFrameCountObjects > maxFrames:
	maxFrames = animationFrameCountObjects

lastFrame = currentFrameNumber + maxFrames + animationFrameCountHoldAtEnd
if fadeOut:
	lastFrame += (animationFrameCountTextFadeout * 1.0)
WriteToFile(dlPath, "Last frame number is %s" % (lastFrame))
bpy.context.scene.frame_start = 0
bpy.context.scene.frame_end = lastFrame
bpy.context.scene.frame_set(0)
WriteToFile(dlPath, "Done assembling")
print("Done assembling")

