#!/usr/bin/env python3

import json
import ntpath
import os
import sys

gLineSeparator = "\n"

def WriteToFile(outputFilePath, content):
	outputFile = open(outputFilePath,"a")
	outputFile.write(content)
	outputFile.write(gLineSeparator)
	outputFile.close()

# https://stackoverflow.com/questions/8384737/extract-file-name-from-path-no-matter-what-the-os-path-format
def pathLeaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)

def showBanner():
	print('Soul Reaver ModelEx Conversion Generation Script')
	print('Ben Lincoln, 2019')
	print('https://www.thelostworlds.net/')
	
def ShowUsage():
	print('Usage: %s --input-list LIST_FILE_PATH --modelex PATH_TO_MODELEX --base-output-dir OUTPUT_DIRECTORY_FOR_COLLADA_FILES --output OUTPUT_BATCH_SCRIPT_PATH [--areas|--objects] [--spectral]' % (sys.argv[0]))
	print(gLineSeparator)
	('Example: %s --input-list area_drm_files.txt --modelex C:\\ModelEx\\ModelEx.exe --base-output-dir C:\\Soul_Reaver\\Models\\Areas --output C:\\Soul_Reaver\\ModelEx_Conversion-Areas.bat --areas' % (sys.argv[0]))
	print(gLineSeparator)

def isAddedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("add8")
	individualAreas.append("aluka7")
	# individualAreas.append("aluka47")	# In the release version, this is the Fire Forge
	individualAreas.append("aluka48")
	individualAreas.append("bonus1")
	individualAreas.append("cathy50")
	individualAreas.append("cathy65")
	individualAreas.append("cathy67")	# deleted in 1999-06-01, then re-added in 1999-06-10
	individualAreas.append("cathy71")
	individualAreas.append("cathy72")
	individualAreas.append("cathy73")
	individualAreas.append("cathy74")
	individualAreas.append("cathy75")
	individualAreas.append("cathy76")
	individualAreas.append("chrono3")
	individualAreas.append("chrono4")
	individualAreas.append("chrono5")
	individualAreas.append("chrono6")
	individualAreas.append("chrono7")
	individualAreas.append("chrono8")
	individualAreas.append("chrono9")
	individualAreas.append("chrono10")
	individualAreas.append("chrono11")
	individualAreas.append("chrono12")
	individualAreas.append("chrono13")
	individualAreas.append("chrono14")
	individualAreas.append("chrono15")
	individualAreas.append("chrono16")
	individualAreas.append("chrono17")
	individualAreas.append("chrono18")
	individualAreas.append("chrono19")
	individualAreas.append("city19")
	individualAreas.append("city20")
	individualAreas.append("city21")
	individualAreas.append("clfpil4")
	individualAreas.append("fire8")
	individualAreas.append("huba18")
	individualAreas.append("huba19")
	individualAreas.append("huba20")
	individualAreas.append("mrlock14")
	individualAreas.append("nighta10")
	individualAreas.append("nighta11")
	individualAreas.append("nighta12")
	individualAreas.append("nighta13")
	individualAreas.append("nightb10")
	individualAreas.append("nightb11")
	individualAreas.append("oracle23")
	individualAreas.append("oracle24")
	individualAreas.append("out13")
	individualAreas.append("out14")
	# individualAreas.append("out15") # This is the Force Glyph room, renamed from aluka26
	individualAreas.append("out16")
	individualAreas.append("pillars10")
	individualAreas.append("skinnr18")
	individualAreas.append("soundg6")
	individualAreas.append("tower12")
	
	if areaNameLower in individualAreas:
		return True

	return False

def isMovedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("mrlock7")	# moved to the citadel
	individualAreas.append("mrlock8")	# moved to the citadel
	
	if areaNameLower in individualAreas:
		return True

	return False

def isSignificantlyChangedArea(areaName):
	areaNameLower = areaName.lower()
	individualAreas = []
	individualAreas.append("cathy19")	# there is a "cathy19" in the final version, but it's a shadow of what it was in the alpha builds
	individualAreas.append("pillars8")	# pillars8 in the alpha builds connects pillars9 to pillars7 (the alternate entrance to the Sanctuary)
	if areaNameLower in individualAreas:
		return True

	return False

def isDeletedArea(areaName):
	areaNameLower = areaName.lower()
	if areaNameLower.startswith("adda"):
		return True
	if areaNameLower.startswith("lair"):
		return True
	if areaNameLower.startswith("movie"):
		return True
	if areaNameLower.startswith("push"):
		return True
	if areaNameLower.startswith("retreat"):
		return True
	if areaNameLower.startswith("test"):
		return True
	if areaNameLower.startswith("undrct"):
		return True
	
	individualAreas = []
	individualAreas.append("aluka7")
	individualAreas.append("aluka13")
	#individualAreas.append("aluka16") # merged with adjacent area, not actually deleted
	individualAreas.append("aluka17")
	individualAreas.append("aluka18")
	#individualAreas.append("aluka26")	# This is the Force Glyph room, which became out15
	individualAreas.append("aluka35")	# this really is a deleted area, but it's just a short connecting passage
	individualAreas.append("aluka42")
	individualAreas.append("aluka47")	# deleted, then re-added
	individualAreas.append("cathy14")
	individualAreas.append("cathy16")
	individualAreas.append("cathy19")
	individualAreas.append("cathy20")
	individualAreas.append("cathy23")
	individualAreas.append("cathy26")
	individualAreas.append("cathy29")
	individualAreas.append("cathy34")
	individualAreas.append("cathy35")
	individualAreas.append("cathy41")
	individualAreas.append("cathy43")
	individualAreas.append("cathy44")
	individualAreas.append("cathy56")
	individualAreas.append("cathy63")
	individualAreas.append("cathy67")	# deleted in 1999-06-01, then re-added in 1999-06-10
	individualAreas.append("city10")
	individualAreas.append("city11")
	individualAreas.append("city12")
	individualAreas.append("city16")
	individualAreas.append("fill5")
	individualAreas.append("forge6")
	#individualAreas.append("huba12") # Fire Forge, moved to Drowned Abbey (where the Water Forge was), not actually deleted
	individualAreas.append("huba15") # merged with adjacent area, not actually deleted
	individualAreas.append("mrlock3")
	individualAreas.append("mrlock4")
	individualAreas.append("mrlock5")
	individualAreas.append("mrlock6")
	individualAreas.append("mrlock9")
	individualAreas.append("mrlock10")
	individualAreas.append("mrlock11")
	individualAreas.append("mrlock12")
	individualAreas.append("mrlock13")
	individualAreas.append("nighta9")
	individualAreas.append("nightb7")
	individualAreas.append("pillars7")
	individualAreas.append("soundg1")
	individualAreas.append("soundg2")
	#individualAreas.append("stone12")	# stone12 isn't present in later builds, but it was merged into one of the other Stone Glyph areas
	individualAreas.append("train10")

	if areaNameLower in individualAreas:
		return True

	return False

if __name__=='__main__':
	if len(sys.argv) < 3:
		print("Usage: %s FILE_LIST_FILE PATH_TO_MODELEX BASE_OUTPUT_DIRECTORY OUTPUT_FILE" % (sys.argv[0]))
		sys.exit(1)
	
	inputListFilePath = ""
	modelexFilePath = ""
	baseOutputDirectory = ""
	outputFilePath = ""
	filesAreAreas = True
	spectralMode = False
	deletedVsNotDeletedMode = False
	addedVsAlwaysPresentMode = False

	showBanner()
	if len(sys.argv) < 9:
		ShowUsage()
		sys.exit(1)
	
	args2 = []
	
	platformString = ""
	texturePathString = ""
	
	argNum = 0
	while argNum < len(sys.argv):
		currentArg = sys.argv[argNum]
		foundArg = False
		if argNum < (len(sys.argv) - 1):
			nextArg = sys.argv[argNum + 1]
			if foundArg == False:
				if currentArg == "--input-list":
					foundArg = True
					inputListFilePath = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--modelex":
					foundArg = True
					modelexFilePath = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--base-output-dir":
					foundArg = True
					baseOutputDirectory = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--output":
					foundArg = True
					outputFilePath = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--texture-file":
					foundArg = True
					texturePathString = os.path.abspath(nextArg)
			if foundArg:
				argNum += 2
		if foundArg == False:
			args2.append(currentArg)
			argNum += 1

	for a in args2:
		if a == "--areas":
			filesAreAreas = True
		if a == "--objects":
			filesAreAreas = False
		if a == "--spectral":
			spectralMode = True
		if a == "--dreamcast":
			platformString = " --platform dreamcast"
		if a == "--pc":
			platformString = " --platform pc"
		if a == "--playstation":
			platformString = " --platform playstation"
		if a == "--highlight-deleted":
			deletedVsNotDeletedMode = True
		if a == "--highlight-added":
			addedVsAlwaysPresentMode = True
		if a == "--help" or a == "-h" or a == "-?":
			ShowUsage()
			sys.exit(0)

	if modelexFilePath == "" or outputFilePath == "" or inputListFilePath == "" or baseOutputDirectory == "":
		ShowUsage()
		sys.exit(1)
	
	if addedVsAlwaysPresentMode and deletedVsNotDeletedMode:
		print('Warning: both --highlight-deleted and --highlight-added were specified. A few areas were deleted, then re-added later. They will be highlighted as deleted areas in the output.')
	
	outputFile = open(outputFilePath,"w")
	outputFile.write(gLineSeparator)
	outputFile.close()
	
	#Make --discard-portal-polygons only present for areas
	#Add --spectral option for areas
	additionalModelExFlags = " --infer-alphamask --materials-for-all-flags --adjust-uvs"
	
	if filesAreAreas:
		additionalModelExFlags = "--discard-non-visible --ignore-pflag2-for-terrain --unhide-invisible-textures --missing-palettes-in-greyscale --adjust-uvs"
		if spectralMode:
			#additionalModelExFlags = "--discard-portal-polygons --spectral"
			additionalModelExFlags += " --spectral"
		#else:
			#additionalModelExFlags = "--discard-portal-polygons"
		#	additionalModelExFlags = "--discard-non-visible --ignore-pflag2-for-terrain --unhide-invisible-textures --missing-palettes-in-greyscale"
	
	additionalModelExFlags += platformString
	if texturePathString != "":
		additionalModelExFlags += ' --texture-file "%s"' % (texturePathString)
	
	print("Processing '%s'" % (inputListFilePath))
	
	with open(inputListFilePath) as inputListFile:
		for line in inputListFile:
			currentInputFilePath = line.strip().replace('\r', '').replace('\n', '')
			print("Generating conversion script line for '%s'" % (currentInputFilePath))
			areaFileName = pathLeaf(currentInputFilePath)
			dotPos = areaFileName.index(".")
			areaName = areaFileName[0:dotPos]
			convertedFilePath = os.path.join(baseOutputDirectory, "%s.dae" % (areaName))
			mExFlags = additionalModelExFlags
			if addedVsAlwaysPresentMode:
				addedArea = False
				if isAddedArea(areaName):
					addedArea = True
				if isSignificantlyChangedArea(areaName):
					addedArea = True
				if addedArea:
					mExFlags = additionalModelExFlags + ' --rendermode notextures --ignore-vertex-colours --set-polygon-colour --polygon-colour "0.15, 0.0, 1.0, 0.0"'
				else:
					mExFlags = additionalModelExFlags + ' --rendermode notextures --ignore-vertex-colours --set-polygon-colour --polygon-colour "0.15, 0.73, 0.73, 0.73"'
			if deletedVsNotDeletedMode:
				deletedArea = False
				if isDeletedArea(areaName):
					deletedArea = True
				if isSignificantlyChangedArea(areaName):
					deletedArea = True
				if deletedArea:
					mExFlags = additionalModelExFlags + ' --rendermode notextures --ignore-vertex-colours --set-polygon-colour --polygon-colour "0.15, 1.0, 0.0, 0.0"'
				else:
					if not addedVsAlwaysPresentMode:
						mExFlags = additionalModelExFlags + ' --rendermode notextures --ignore-vertex-colours --set-polygon-colour --polygon-colour "0.15, 0.73, 0.73, 0.73"'
			WriteToFile(outputFilePath, '"%s" --input "%s" --output "%s" %s' % (modelexFilePath, currentInputFilePath, convertedFilePath, mExFlags))