#!/usr/bin/env python

import sys

def GetCommands(levelName, levelDescription, originAreaName, levelRegex, spincamDistance, spincamHeight, halfStrengthLights = False, doubleSidedPolygons = False, useEevee = False):

	commonFlags = " --breadth-first --ignore-raziel --ignore-non-visible --portal-override-list %PORTALOVERRIDE% --place-unlinked-areas --place-objects-simultaneously"
	if halfStrengthLights:
		commonFlags += " --overhead-lights-half-strength"
	if doubleSidedPolygons:
		commonFlags += " --double-sided-polygons"
	if useEevee:
		commonFlags += " --eevee"
	
	originAreaNameString = "%ORIGINAREA%"
	if originAreaName != "":
		originAreaNameString = originAreaName
	levelNameString = ""
	if levelName != "":
		levelNameString = '-%s' % (levelName)
	else:
		commonFlags = "%s --place-unused-objects" % (commonFlags)
	levelNoteString = ""
	if levelDescription != "":
		levelNoteString = " (%s)" % (levelDescription)
		print('REM ######## ISOLATED LEVEL: %s ########' % (levelDescription))
		print()
	levelRegexString = ""
	if levelRegex != "":
		levelRegexString = ' --area-whitelist "%s"' % (levelRegex)
	
	
	spinOptions = "--add-spincam --spin-num-rotations 1 --spin-num-frames 300 --spin-start-height %s --spin-end-height %s --spin-start-distance %s --spin-end-distance %s" % (spincamHeight, spincamHeight, spincamDistance, spincamDistance)

	#   
	tripleSpinOptions = "--add-tt-spincam --tt-spin-distance1 %s --tt-spin-distance2 %s --tt-spin-distance3 %s --tt-spin-height1 %s --tt-spin-height2 %s --tt-spin-height3 %s --spin-num-frames 600" % (spincamDistance, spincamDistance * 1.5, spincamDistance, spincamHeight, 0.0, spincamHeight * -1.0)

	print('REM material plane with objects%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position%s-material.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s.txt" --material --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM material plane with objects%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position%s-material.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s.txt" --material --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()

	print('REM material plane with objects except monsters%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-no_monsters.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-No_Monsters.txt" --material --import-objects --ignore-monsters %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-No_Monsters.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM material plane with objects except monsters%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-no_monsters.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-No_Monsters.txt" --material --import-objects --ignore-monsters %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-No_Monsters.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()

	print('REM material plane without objects%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-no_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-No_Objects.txt" --material %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-No_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM material plane without objects%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-no_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-No_Objects.txt" --material %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-No_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()

	print('REM material plane with all objects (including spectral and non-visible objects)%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-all_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-All_Objects.txt" --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-All_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM material plane with all objects (including spectral and non-visible objects)%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREAS%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-material%s-all_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Material%s-All_Objects.txt" --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Material%s-All_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()
	
	print('REM spectral plane with objects%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position%s-spectral.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s.txt" --spectral --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM spectral plane with objects%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position%s-spectral.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s.txt" --spectral --import-objects %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()

	print('REM spectral plane with objects except monsters%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-spectral%s-no_monsters.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s-No_Monsters.txt" --spectral --import-objects --ignore-monsters %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s-No_Monsters.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM spectral plane with objects except monsters%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-spectral%s-no_monsters.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s-No_Monsters.txt" --spectral --import-objects --ignore-monsters %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s-No_Monsters.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()

	print('REM spectral plane without objects%s (spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-spectral%s-no_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s-No_Objects.txt" --spectral %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s-No_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, spinOptions, levelNameString))
	print()
	
	print('REM spectral plane without objects%s (triple spincam):' % (levelNoteString))
	print('python "%%POSITIONCALCULATIONSCRIPT%%" "%s" --start-area-name %s --start-area-file "%%JSONDIR%%\\%s.json" --areas-collada-dir "%%MODELSDIRAREASSPECTRAL%%" --objects-collada-dir "%%MODELSDIROBJECTS%%" --output "%%BASEPATH%%\\%%BUILDDATE%%\\blender_import_and_position-spectral%s-no_objects.py" --log "%%BASEPATH%%\\%%BUILDDATE%%\\Blender_Import_Debug_Log-Spectral%s-No_Objects.txt" --spectral %s %s --add-overhead-lights %s > "%%BASEPATH%%\\%%BUILDDATE%%\\Position_Calculation_Log-Spectral%s-No_Objects.txt" 2>&1' % (originAreaNameString, originAreaNameString, originAreaNameString, levelNameString, levelNameString, commonFlags, levelRegexString, tripleSpinOptions, levelNameString))
	print()


if __name__=='__main__':
	args2 = []
	
	argNum = 0
	while argNum < len(sys.argv):
		currentArg = sys.argv[argNum]
		foundArg = False
		if argNum < (len(sys.argv) - 1):
			nextArg = sys.argv[argNum + 1]
			# if currentArg == "--start-area-file":
				# foundArg = True
				# inputFilePath = os.path.abspath(nextArg)
			# if foundArg == False:
				# if currentArg == "--start-area-name":
					# foundArg = True
					# startArea = nextArg
			if foundArg:
				argNum += 2
		if foundArg == False:
			args2.append(currentArg)
			argNum += 1

	halfStrengthLights = False
	doubleSidedPolygons = False

	for a in args2:
		if a == "--overhead-lights-half-strength":	
			halfStrengthLights = True
		if a == "--double-sided-polygons":	
			doubleSidedPolygons = True
		if a == "--eevee":
			useEevee = True
		#if a == "--help" or a == "-h" or a == "-?":
		#	ShowUsage()
		#	sys.exit(0)
			
	GetCommands("", "", "", "", 400, 350, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("Underworld", "Underworld", "under1", "(under|train)[0-9]+", 75, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("PillarsTombAlpha", "The Pillars of Nosgoth and the Sarafan Tomb (alpha builds)", "add8", "(add|bonus|boss|conect|connect|conecta|conectb|conectc|conectd|conecte|conectf|forge|pillar|pillars|soundg|tomb|tompil)[0-9]+", 200, 170, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("PillarsTombBetaRelease", "The Pillars of Nosgoth and the Sarafan Tomb (beta and release builds)", "add8", "(add|bonus|boss|conect|connect|conecta|conectb|conectc|conectd|conecte|conectf|pillar|pillars|tomb|tompil)[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("LakeOfTheDead", "The Lake of the Dead", "cliff1", "(cliff|clfpil|clfsubb)[0-9]+", 75, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("MelchiahimClanTerritorySkinnr5", "Melchiah's Clan Territory (Starting Layout from Skinnr5)", "skinnr5", "(skinnr|out)[0-9]+", 150, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("MelchiahimClanTerritorySkinnr1", "Melchiah's Clan Territory (Starting Layout from Skinnr1)", "skinnr1", "(skinnr|out)[0-9]+", 150, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("MelchiahimClanTerritoryOut1", "Melchiah's Clan Territory (Starting Layout from Out1)", "skinnr1", "(skinnr|out)[0-9]+", 150, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("SilencedCathedralAlpha", "Silenced Cathedral / Zephon's Clan Territory (alpha builds)", "cathy1", "cathy[0-9]+", 200, 200, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("SilencedCathedralBetaRelease", "Silenced Cathedral / Zephon's Clan Territory (beta and release builds)", "cathy1", "(?!(cathy50|cathy58))((cathy|soundg)[0-9]+)", 200, 200, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	# 200 distance is too much for the Drowned Abbey
	GetCommands("DrownedAbbey", "DrownedAbbey / Rahab's Clan Territory", "aluka2", "(aluka|chapel|conectg)[0-9]+", 125, 150, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("AshVillage", "Ash Village / Dumah's Clan Territory", "nighta1", "(adda|hubb|nighta|nightb)[0-9]+", 125, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	# 200 distance is WAY too much for Turel's territory
	GetCommands("TurelimClanTerritory", "Turel's Clan Territory (alpha builds)", "lair1", "(mrlock|lair)[0-9]+", 50, 125, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("RazielimClanTerritory", "Raziel's Clan Territory", "huba1", "huba[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("Chronoplast", "Chronoplast / Oracle's Cave", "oracle1", "(oracle|chrono)[0-9]+", 100, 50, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("CitadelAlpha", "Human Citadel (alpha builds)", "city1", "(city|cityout|tower|undrct)[0-9]+", 200, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("CitadelBetaRelease", "Human Citadel (beta and release builds)", "city1", "(city|cityout|tower|mrlock)[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("FireGlyph", "Fire Glyph", "fire1", "fire[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons)
	GetCommands("StoneGlyph", "Stone Glyph", "skinnr1", "(stone|out)[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("SunlightGlyph", "Sunlight Glyph", "intvaly1", "(fill|filtuna|filtunb|gasrm|gastuna|gastunb|htorm|htotun|intvaly|litein|litetun|litrnda|litrndb|pisaira|pisairb|pistair|piston|pistunl|sunrm|suntnl|suntnlb)[0-9]+", 200, 50, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("MountainRetreat", "Kain's Mountain Retreat (alpha builds)", "retreat1", "retreat[0-9]+", 75, 75, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("MovieRooms", "Movie Rooms (1999-06-01 and 1999-06-10)", "push1", "(push1|movie[0-9]+)", 150, 50, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("TestAreas", "Test Areas", "cliff1", "(cliff|push|test)[0-9]+", 300, 50, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)
	GetCommands("DemoAreas", "Demo Areas", "demo1", "demo[0-9]+", 100, 100, halfStrengthLights = halfStrengthLights, doubleSidedPolygons = doubleSidedPolygons, useEevee = useEevee)



