import bpy
import re

def getNodeByName(material, nodeName):
	for n in material.node_tree.nodes:
		if n.name == nodeName:
			return n
	return None

def getObjectWithShortestName(objectList):
	minLength = 10000
	for o in objectList:
		if len(o.name) < minLength:
			minLength = len(o.name)
	for o in objectList:
		if len(o.name) == minLength:
			return o
	return None

def getImportedTopLevelObjects(objectList):
	result = []
	for o in objectList:
		if not o.parent or o.parent == '':
			result.append(o)
	return result

def renameChildObjects(objectList, oldString, newString):
	for o in objectList:
		if o.name != oldString and oldString in o.name:
			o.name = o.name.replace(oldString, newString)
			if o.data:
				o.data.name = o.data.name.replace(oldString, newString)

# loosely based on https://blenderartists.org/t/question-regarding-expanding-collapsing-collection-in-outliner-in-2-8/1175242
# the original version of the code only expanded/collapsed the outliner in the Scripting view
# this version applies the changes to all views
def toggle_expand(state):
	for s in bpy.data.screens.keys():
		bpy.ops.wm.redraw_timer(type='DRAW_WIN', iterations=1)
		for a in bpy.data.screens[s].areas:
			if a.type == 'OUTLINER':
				bpy.ops.outliner.show_hierarchy({'area': a}, 'INVOKE_DEFAULT')
				for i in range(state):
					bpy.ops.outliner.expanded_toggle({'area': a})
				a.tag_redraw()

def adjustMaterials():
	for mat in bpy.data.materials:
		# enable backface culling (single-sided polygons)
		mat.use_backface_culling = True
		# disable display of transparent backs, because it causes visual corruption when backface culling is enabled
		mat.show_transparent_back = False
		opacity = 1.0
		if 'trans' in mat.name:
			#opacity_string = re.search('opacity_[0-9\.,_]+_$', mat.name).group()
			opacity_string = re.search('opacity_[0-9_]+_($|.)', mat.name).group()
			if opacity_string != '':
				# enable alpha blend mode (for transparent/translucent textures)
				opacity_string = opacity_string[8:]
				opacity_string = opacity_string[:-1]
				opacity_string = opacity_string.replace("_", ".")
				try:
					opacity = float(opacity_string)
				except Exception as e:
					print("Couldn't convert '%s' to a floating-point number" % (opacity_string))
					# handle locales that use comma as a decimal point instead of a period
					opacity_string = opacity_string.replace(",", ".")
					try:
						opacity = float(opacity_string)
					except Exception as e:
						print("Couldn't convert '%s' to a floating-point number" % (opacity_string))
				if opacity < 1.0:
					mat.blend_method = 'BLEND'
		else:
			#if "alphamask" in mat.name:
				# enable alpha clipping mode (for opaque extures with some 100% transparent parts)
			mat.blend_method = 'CLIP'
		# link texture alpha to BSDF alpha
		mat.use_nodes= True
		bsdfNode = None
		baseColourNode = None
		mixShaderNode = None
		vertexColourNode = None
		valueNode = None
		for n in mat.node_tree.nodes:
			print(n.name)
			if n.name == 'Principled BSDF':
				bsdfNode = n
			if n.name == 'Image Texture':
				baseColourNode = n
			if n.name == 'Vertex Color':
				vertexColourNode = n
			if n.name == 'Mix':
				mixShaderNode = n
			if n.name == 'Value':
				valueNode = n
		if bsdfNode and baseColourNode:
			if not mixShaderNode or not vertexColourNode:
				if opacity < 1.0:
					mat.node_tree.nodes.new(type = "ShaderNodeValue")
					valueNode = getNodeByName(mat, "Value")
					valueNode.outputs[0].default_value = opacity
					mat.node_tree.links.new(valueNode.outputs[0], bsdfNode.inputs['Alpha'])
				else:
					mat.node_tree.links.new(baseColourNode.outputs['Alpha'], bsdfNode.inputs['Alpha'])
				mat.node_tree.nodes.new(type = "ShaderNodeMixRGB")
				mixShaderNode = getNodeByName(mat, "Mix")
				mixShaderNode.inputs['Fac'].default_value = 1.0
				mixShaderNode.blend_type = 'COLOR'
				mat.node_tree.nodes.new(type = "ShaderNodeVertexColor")
				vertexColourNode = getNodeByName(mat, "Vertex Color")
				# Remove existing links, where necessary
				for existingLink in baseColourNode.outputs['Color'].links:
					mat.node_tree.links.remove(existingLink)
				mat.node_tree.links.new(baseColourNode.outputs['Color'], mixShaderNode.inputs['Color1'])
				mat.node_tree.links.new(vertexColourNode.outputs['Color'], mixShaderNode.inputs['Color2'])
				mat.node_tree.links.new(mixShaderNode.outputs['Color'], bsdfNode.inputs['Base Color'])
				# set specular to 0 to more accurately reflect the game
				bsdfNode.inputs['Specular'].default_value = 0
				
	print("Max distance from origin: %s" % (pTracker.MaxDistanceFromOrigin))

	#post-import steps
	WriteToFile(outputFilePath, "# collapse all collection entries in all contexts/views for convenience")
	WriteToFile(outputFilePath, "toggle_expand(2)")
	WriteToFile(outputFilePath, "# adjust materials for all objects")
	WriteToFile(outputFilePath, "adjustMaterials()")